@@ -91,8 +91,7 @@ static int set_assoc_params(struct wrap_ndis_device *wnd)
 		else
 			auth_mode = Ndis802_11AuthModeWPANone;
 	} else if (wnd->iw_auth_80211_auth_alg & IW_AUTH_ALG_SHARED_KEY) {
-		if (wnd->iw_auth_80211_auth_alg &
-		    IW_AUTH_ALG_OPEN_SYSTEM)
+		if (wnd->iw_auth_80211_auth_alg & IW_AUTH_ALG_OPEN_SYSTEM)
 			auth_mode = Ndis802_11AuthModeAutoSwitch;
 		else
 			auth_mode = Ndis802_11AuthModeShared;
@@ -695,8 +694,7 @@ int set_auth_mode(struct wrap_ndis_device *wnd, int auth_mode)
 	NDIS_STATUS res;
 
 	TRACEENTER2("%d", auth_mode);
-	res = miniport_set_int(wnd, OID_802_11_AUTHENTICATION_MODE,
-			       auth_mode);
+	res = miniport_set_int(wnd, OID_802_11_AUTHENTICATION_MODE, auth_mode);
 	if (res == NDIS_STATUS_FAILURE)
 		return -EOPNOTSUPP;
 	if (res == NDIS_STATUS_INVALID_DATA) {
@@ -891,17 +889,22 @@ static int remove_key(struct wrap_ndis_device *wnd, int index,
 		struct ndis_remove_key remove_key;
 		remove_key.struct_size = sizeof(remove_key);
 		remove_key.index = index;
-		/* pairwise key */
-		if (memcmp(bssid, "\xff\xff\xff\xff\xff\xff", ETH_ALEN) != 0)
-			remove_key.index |= (1 << 30);
-		memcpy(remove_key.bssid, bssid, sizeof(remove_key.bssid));
+		if (bssid) {
+			/* pairwise key */
+			if (memcmp(bssid, "\xff\xff\xff\xff\xff\xff",
+				   ETH_ALEN) != 0)
+				remove_key.index |= (1 << 30);
+			memcpy(remove_key.bssid, bssid,
+			       sizeof(remove_key.bssid));
+		} else
+			memset(remove_key.bssid, 0xff,
+			       sizeof(remove_key.bssid));
 		if (miniport_set_info(wnd, OID_802_11_REMOVE_KEY,
 				      (char *)&remove_key, sizeof(remove_key)))
 			TRACEEXIT2(return -EINVAL);
 	} else {
 		ndis_key_index keyindex = index;
-		res = miniport_set_info(wnd, OID_802_11_REMOVE_WEP,
-					&keyindex, sizeof(keyindex));
+		res = miniport_set_int(wnd, OID_802_11_REMOVE_WEP, keyindex);
 		if (res == NDIS_STATUS_FAILURE)
 			TRACEEXIT2(return -EOPNOTSUPP);
 		if (res == NDIS_STATUS_INVALID_DATA) {
@@ -1110,8 +1113,7 @@ static char *ndis_translate_scan(struct net_device *dev, char *event,
 		iwe.u.data.flags = IW_ENCODE_ENABLED | IW_ENCODE_NOKEY;
 	iwe.u.data.length = 0;
 	iwe.len = IW_EV_POINT_LEN;
-	event = iwe_stream_add_point(event, end_buf, &iwe,
-				     bssid->ssid.essid);
+	event = iwe_stream_add_point(event, end_buf, &iwe, bssid->ssid.essid);
 
 	/* add rate */
 	memset(&iwe, 0, sizeof(iwe));
@@ -1150,19 +1152,7 @@ static char *ndis_translate_scan(struct net_device *dev, char *event,
 	if (bssid->length > sizeof(*bssid)) {
 		unsigned char *iep = (unsigned char *)bssid_ex->ies +
 			sizeof(struct ndis_fixed_ies);
-		unsigned char *end = iep + bssid_ex->ie_length;
-	/*
-	 * TODO: backwards compatibility would require that IWEVCUSTOM
-	 * is send even if WIRELESS_EXT > 17. This version does not do
-	 * this in order to allow wpa_supplicant to be tested with
-	 * WE-18.
-	 */
-#if 0
-		memset(&iwe, 0, sizeof(iwe));
-		iwe.cmd = IWEVGENIE;
-		iwe.u.data.length = bssid_ex->ie_length;
-		event = iwe_stream_add_point(event, end_buf, &iwe, iep);
-#endif
+		__unused unsigned char *end = iep + bssid_ex->ie_length;
 
 		while (iep + 1 < end && iep + 2 + iep[1] <= end) {
 			unsigned char ielen = 2 + iep[1];
@@ -1171,6 +1161,17 @@ static char *ndis_translate_scan(struct net_device *dev, char *event,
 				iep += ielen;
 				continue;
 			}
+#if WIRELESS_EXT > 17
+			if ((iep[0] == WLAN_EID_GENERIC && iep[1] >= 4 &&
+			     memcmp(iep + 2, "\x00\x50\xf2\x01", 4) == 0) ||
+			    iep[0] == RSN_INFO_ELEM) {
+				memset(&iwe, 0, sizeof(iwe));
+				iwe.cmd = IWEVGENIE;
+				iwe.u.data.length = ielen;
+				event = iwe_stream_add_point(event, end_buf,
+							     &iwe, iep);
+			}
+#else
 
 			if (iep[0] == WLAN_EID_GENERIC && iep[1] >= 4 &&
 			    memcmp(iep + 2, "\x00\x50\xf2\x01", 4) == 0) {
@@ -1203,11 +1204,10 @@ static char *ndis_translate_scan(struct net_device *dev, char *event,
 				event = iwe_stream_add_point(event, end_buf,
 							     &iwe, buf);
 			}
-
+#endif
 			iep += ielen;
 		}
 	}
-
 	DBGTRACE2("event = %p, current_val = %p", event, current_val);
 	TRACEEXIT2(return event);
 }
@@ -1217,17 +1217,14 @@ int set_scan(struct wrap_ndis_device *wnd)
 	NDIS_STATUS res;
 
 	TRACEENTER2("");
-	res = miniport_set_int(wnd, OID_802_11_BSSID_LIST_SCAN, 0);
-	if (res == NDIS_STATUS_FAILURE)
-		return -EOPNOTSUPP;
-	if (res == NDIS_STATUS_NOT_SUPPORTED ||
+	res = miniport_set_info(wnd, OID_802_11_BSSID_LIST_SCAN, NULL, 0);
+	if (res == NDIS_STATUS_FAILURE || res == NDIS_STATUS_NOT_SUPPORTED ||
 	    res == NDIS_STATUS_INVALID_DATA) {
 		WARNING("scanning failed (%08X)", res);
 		TRACEEXIT2(return -EOPNOTSUPP);
-	} else {
-		wnd->scan_timestamp = jiffies;
-		TRACEEXIT2(return 0);
 	}
+	wnd->scan_timestamp = jiffies;
+	TRACEEXIT2(return 0);
 }
 
 static int iw_set_scan(struct net_device *dev, struct iw_request_info *info,
@@ -1243,7 +1240,7 @@ static int iw_get_scan(struct net_device *dev, struct iw_request_info *info,
 	struct wrap_ndis_device *wnd = netdev_priv(dev);
 	unsigned int i, list_len, needed;
 	NDIS_STATUS res;
-	struct ndis_bssid_list *bssid_list;
+	struct ndis_bssid_list *bssid_list = NULL;
 	char *event = extra;
 	struct ndis_wlan_bssid *cur_item ;
 
@@ -1264,7 +1261,7 @@ static int iw_get_scan(struct net_device *dev, struct iw_request_info *info,
 	needed = 0;
 	res = miniport_query_info_needed(wnd, OID_802_11_BSSID_LIST,
 					 bssid_list, list_len, &needed);
-	if (needed > 0 || res == NDIS_STATUS_INVALID_LENGTH ||
+	if (res == NDIS_STATUS_INVALID_LENGTH ||
 	    res == NDIS_STATUS_BUFFER_TOO_SHORT) {
 		/* now try with required space */
 		kfree(bssid_list);
@@ -1279,24 +1276,21 @@ static int iw_get_scan(struct net_device *dev, struct iw_request_info *info,
 		res = miniport_query_info(wnd, OID_802_11_BSSID_LIST,
 					  bssid_list, list_len);
 	}
-
-	if (res == NDIS_STATUS_INVALID_DATA || res == NDIS_STATUS_FAILURE) {
+	if (res) {
 		WARNING("getting BSSID list failed (%08X)", res);
 		kfree(bssid_list);
 		TRACEEXIT2(return -EOPNOTSUPP);
 	}
-
-	for (i = 0, cur_item = &bssid_list->bssid[0] ;
-	     i < bssid_list->num_items ; i++) {
-		event = ndis_translate_scan(dev, event,
-					    extra + IW_SCAN_MAX_DATA,
+	DBGTRACE2("%d", bssid_list->num_items);
+	cur_item = &bssid_list->bssid[0];
+	for (i = 0; i < bssid_list->num_items; i++) {
+		event = ndis_translate_scan(dev, event, extra + IW_SCAN_MAX_DATA,
 					    cur_item);
 		cur_item = (struct ndis_wlan_bssid *)((char *)cur_item +
 						      cur_item->length);
 	}
 	wrqu->data.length = event - extra;
 	wrqu->data.flags = 0;
-
 	kfree(bssid_list);
 	TRACEEXIT2(return 0);
 }
@@ -1460,14 +1454,14 @@ static int iw_get_range(struct net_device *dev, struct iw_request_info *info,
 	range->encoding_size[1] = 13;
 
 	range->num_bitrates = 0;
+	n = NDIS_MAX_RATES_EX;
 	res = miniport_query_info(wnd, OID_802_11_SUPPORTED_RATES,
 				  &rates_ex, sizeof(ndis_rates_ex));
 	if (res) {
 		res = miniport_query_info(wnd, OID_802_11_SUPPORTED_RATES,
 					  &rates_ex, sizeof(ndis_rates));
 		n = NDIS_MAX_RATES;
-	} else
-		n = NDIS_MAX_RATES_EX;
+	}
 	if (res)
 		WARNING("getting bit rates failed: %08X", res);
 	else {
@@ -1523,6 +1517,7 @@ static int iw_get_range(struct net_device *dev, struct iw_request_info *info,
 	if (test_bit(Ndis802_11AuthModeWPA2, &wnd->capa.auth) ||
 	    test_bit(Ndis802_11AuthModeWPA2PSK, &wnd->capa.auth))
 		range->enc_capa |= IW_ENC_CAPA_WPA2;
+
 #endif /* WIRELESS_EXT > 17 */
 
 	return 0;
@@ -1532,6 +1527,9 @@ static int iw_get_range(struct net_device *dev, struct iw_request_info *info,
 static int wpa_disassociate(struct net_device *dev,
 			    struct iw_request_info *info,
 			    union iwreq_data *wrqu, char *extra);
+static int wpa_deauthenticate(struct net_device *dev,
+			      struct iw_request_info *info,
+			      union iwreq_data *wrqu, char *extra);
 
 static int iw_set_mlme(struct net_device *dev, struct iw_request_info *info,
 		       union iwreq_data *wrqu, char *extra)
@@ -1541,6 +1539,7 @@ static int iw_set_mlme(struct net_device *dev, struct iw_request_info *info,
 	TRACEENTER2("");
 	switch (mlme->cmd) {
 	case IW_MLME_DEAUTH:
+		return wpa_deauthenticate(dev, info, wrqu, extra);
 	case IW_MLME_DISASSOC:
 		DBGTRACE2("cmd=%d reason_code=%d",
 			  mlme->cmd, mlme->reason_code);
@@ -1589,9 +1588,9 @@ static int iw_set_auth(struct net_device *dev,
 	case IW_AUTH_80211_AUTH_ALG:
 		wnd->iw_auth_80211_auth_alg = wrqu->param.value;
 		break;
+	case IW_AUTH_WPA_ENABLED:
 	case IW_AUTH_TKIP_COUNTERMEASURES:
 	case IW_AUTH_DROP_UNENCRYPTED:
-	case IW_AUTH_WPA_ENABLED:
 	case IW_AUTH_RX_UNENCRYPTED_EAPOL:
 	case IW_AUTH_PRIVACY_INVOKED:
 		/* TODO */
@@ -1779,8 +1778,7 @@ static int iw_set_pmksa(struct net_device *dev,
 	pmkid.length = 8 + pmkid.bssid_info_count *
 		sizeof(struct ndis_bssid_info);
 
-	res = miniport_set_info(wnd, OID_802_11_PMKID, &pmkid,
-				sizeof(pmkid));
+	res = miniport_set_info(wnd, OID_802_11_PMKID, &pmkid, sizeof(pmkid));
 	if (res == NDIS_STATUS_FAILURE)
 		return -EOPNOTSUPP;
 	DBGTRACE2("OID_802_11_PMKID -> %d", res);
@@ -2145,10 +2143,10 @@ static int wpa_disassociate(struct net_device *dev,
 			    union iwreq_data *wrqu, char *extra)
 {
 	struct wrap_ndis_device *wnd = netdev_priv(dev);
-	unsigned char buf[NDIS_ESSID_MAX_SIZE];
+	u8 buf[NDIS_ESSID_MAX_SIZE];
 	int i;
 
-	TRACEENTER2("");
+	miniport_set_info(wnd, OID_802_11_DISASSOCIATE, NULL, 0);
 	get_random_bytes(buf, sizeof(buf));
 	for (i = 0; i < sizeof(buf); i++)
 		buf[i] = 'a' + (buf[i] % 26);
@@ -2279,10 +2277,14 @@ static int wpa_deauthenticate(struct net_device *dev,
 			      struct iw_request_info *info,
 			      union iwreq_data *wrqu, char *extra)
 {
+	struct wrap_ndis_device *wnd = netdev_priv(dev);
 	int ret;
 
 	TRACEENTER2("");
 	ret = wpa_disassociate(dev, info, wrqu, extra);
+	set_priv_filter(wnd, Ndis802_11PrivFilterAcceptAll);
+	set_auth_mode(wnd, Ndis802_11AuthModeOpen);
+	set_encr_mode(wnd, Ndis802_11EncryptionDisabled);
 	TRACEEXIT2(return ret);
 }
 
