@@ -372,8 +372,7 @@ static int load_settings(struct wrap_driver *wrap_driver,
 	}
 	nt_spin_unlock(&loader_lock);
 	if (!wd) {
-		ERROR("conf file %s not found",
-		      wrap_devices[i].conf_file_name);
+		ERROR("conf file %s not found", wrap_devices[i].conf_file_name);
 		TRACEEXIT1(return -EINVAL);
 	}
 
@@ -409,8 +408,8 @@ static int load_settings(struct wrap_driver *wrap_driver,
 					WRAP_DEVICE_BUS_TYPE(dev_type,
 							     bus_type);
 			DBGTRACE2("data1: %x, dev type: %x, bus type: %x, "
-				  "new: %x\n",
-				  data1, dev_type, bus_type, wd->dev_bus_type);
+				  "new: %x\n", data1, dev_type, bus_type,
+				  wd->dev_bus_type);
 		}
 		nt_spin_lock(&loader_lock);
 		InsertTailList(&wd->settings, &setting->list);
@@ -869,6 +868,13 @@ void loader_exit(void)
 	TRACEENTER1("");
 	misc_deregister(&wrapper_misc);
 
+	/* mark so device's remove function knows it is called because
+	 * module is being removed (and not because device is
+	 * unplugged) */
+	/* TODO: is there a better way to infer this? */
+	for (i = 0; i < num_wrap_devices; i++)
+		wrap_devices[i].surprise_removed = FALSE;
+
 #ifdef CONFIG_USB
 	if (wrap_usb_devices) {
 		usb_deregister(&wrap_usb_driver);
