@@ -59,7 +59,7 @@ void ndis_exit(void)
 void ndis_exit_device(struct wrap_ndis_device *wnd)
 {
 	struct wrap_device_setting *setting;
-
+	DBGTRACE2("%p", wnd);
 	/* TI driver doesn't call NdisMDeregisterInterrupt during halt! */
 	if (wnd->ndis_irq)
 		NdisMDeregisterInterrupt(wnd->ndis_irq);
@@ -389,6 +389,8 @@ STDCALL ULONG WRAP_EXPORT(NDIS_BUFFER_TO_SPAN_PAGES)
 		return 1;
 	length = MmGetMdlByteCount(buffer);
 	n = SPAN_PAGES(MmGetMdlVirtualAddress(buffer), length);
+	DBGTRACE4("%p, %p, %d, %d", buffer->startva, buffer->mappedsystemva,
+		  length, n);
 	TRACEEXIT3(return n);
 }
 
@@ -413,15 +415,13 @@ ndis_encode_setting(struct wrap_device_setting *setting,
 			return param;
 		if (param->type == NdisParameterString)
 			RtlFreeUnicodeString(&param->data.string);
-		ExFreePool(param);
 		setting->encoded = NULL;
-	}
-	param = ExAllocatePoolWithTag(NonPagedPool, sizeof(*param), 0);
+	} else
+		param = ExAllocatePoolWithTag(NonPagedPool, sizeof(*param), 0);
 	if (!param) {
 		ERROR("couldn't allocate memory");
 		return NULL;
 	}
-	param->type = type;
 	TRACEENTER2("type = %d", type);
 	switch(type) {
 	case NdisParameterInteger:
@@ -446,6 +446,7 @@ ndis_encode_setting(struct wrap_device_setting *setting,
 		ExFreePool(param);
 		return NULL;
 	}
+	param->type = type;
 	setting->encoded = param;
 	TRACEEXIT2(return param);
 }
@@ -458,13 +459,11 @@ static int ndis_decode_setting(struct wrap_device_setting *setting,
 	TRACEENTER2("%p, %p", setting, param);
 	switch(param->type) {
 	case NdisParameterInteger:
-		snprintf(setting->value, sizeof(u32), "%u",
-			 param->data.integer);
+		snprintf(setting->value, sizeof(u32), "%u", param->data.integer);
 		setting->value[sizeof(ULONG)] = 0;
 		break;
 	case NdisParameterHexInteger:
-		snprintf(setting->value, sizeof(u32), "%x",
-			 param->data.integer);
+		snprintf(setting->value, sizeof(u32), "%x", param->data.integer);
 		setting->value[sizeof(ULONG)] = 0;
 		break;
 	case NdisParameterString:
@@ -659,16 +658,7 @@ STDCALL void WRAP_EXPORT(NdisMSetAttributesEx)
 	if (attributes & NDIS_ATTRIBUTE_BUS_MASTER)
 		pci_set_master(wnd->wd->pci.pdev);
 
-	if (!(attributes & NDIS_ATTRIBUTE_DESERIALIZE)) {
-		DBGTRACE2("serialized driver");
-		set_bit(ATTR_SERIALIZED, &wnd->attributes);
-	}
-
-	if (attributes & NDIS_ATTRIBUTE_SURPRISE_REMOVE_OK)
-		set_bit(ATTR_SURPRISE_REMOVE, &wnd->attributes);
-
-	if (attributes & NDIS_ATTRIBUTE_NO_HALT_ON_SUSPEND)
-		set_bit(ATTR_NO_HALT_ON_SUSPEND, &wnd->attributes);
+	wnd->attributes = attributes;
 
 	if (hangcheck_interval > 0)
 		wnd->hangcheck_interval = 2 * hangcheck_interval * HZ;
@@ -919,10 +909,9 @@ STDCALL void WRAP_EXPORT(NdisMAllocateSharedMemory)
 		    nmb->wnd->dma_map_count, size, cached);
 
 	*virt = PCI_DMA_ALLOC_COHERENT(wd->pci.pdev, size, &dma_addr);
-	if (!*virt) {
+	if (!*virt)
 		ERROR("couldn't allocate %d bytes of %scached DMA memory",
 		      size, cached ? "" : "un-");
-	}
 	*phys = dma_addr;
 	TRACEEXIT3(return);
 }
@@ -991,7 +980,7 @@ STDCALL NDIS_STATUS WRAP_EXPORT(NdisMAllocateSharedMemoryAsync)
 	alloc_shared_mem->cached = cached;
 	alloc_shared_mem->ctx = ctx;
 	if (schedule_ntos_work_item(alloc_shared_memory_async,
-				    wnd, alloc_shared_mem, FALSE))
+				    wnd, alloc_shared_mem, WORKER_FUNC_LINUX))
 		TRACEEXIT3(return NDIS_STATUS_FAILURE);
 	TRACEEXIT3(return NDIS_STATUS_PENDING);
 }
@@ -1082,6 +1071,8 @@ STDCALL void WRAP_EXPORT(NdisAllocateBuffer)
 		irql = nt_spin_lock_irql(&pool->lock, DISPATCH_LEVEL);
 		pool->num_allocated_descr++;
 	}
+//	descr->flags |= MDL_PAGES_LOCKED | MDL_MAPPED_TO_SYSTEM_VA;
+//	descr->flags |= MDL_SOURCE_IS_NONPAGED_POOL | MDL_ALLOCATED_FIXED_SIZE;
 	MmBuildMdlForNonPagedPool(descr);
 	/* NdisFreeBuffer doesn't pass pool, so we store pool
 	 * in unused field 'process' */
@@ -1480,7 +1471,9 @@ STDCALL void WRAP_EXPORT(NdisAllocatePacket)
 	memset(ndis_packet, 0, packet_length);
 	ndis_packet->private.oob_offset = packet_length -
 		sizeof(struct ndis_packet_oob_data);
-	ndis_packet->private.packet_flags = fPACKET_ALLOCATED_BY_NDIS;
+	ndis_packet->private.packet_flags =
+		fPACKET_ALLOCATED_BY_NDIS;
+//		fPACKET_ALLOCATED_BY_NDIS | NDIS_PROTOCOL_ID_TCP_IP;
 	ndis_packet->private.pool = pool;
 	nt_spin_unlock_irql(&pool->lock, irql);
 
@@ -1628,7 +1621,7 @@ STDCALL void WRAP_EXPORT(NdisSend)
 		LIN2WIN3(miniport->send_packets, wnd->nmb->adapter_ctx,
 			 packets, 1);
 		lower_irql(irql);
-		if (test_bit(ATTR_SERIALIZED, &wnd->attributes)) {
+		if (!(wnd->attributes & NDIS_ATTRIBUTE_DESERIALIZE)) {
 			struct ndis_packet_oob_data *oob_data;
 			oob_data = NDIS_PACKET_OOB_DATA(packet);
 			*status = oob_data->status;
@@ -1803,8 +1796,7 @@ static void ndis_irq_handler(unsigned long data)
 	miniport = &wnd->wd->driver->ndis_driver->miniport;
 	LIN2WIN1(miniport->handle_interrupt, wnd->nmb->adapter_ctx);
 	if (miniport->enable_interrupts)
-		LIN2WIN1(miniport->enable_interrupts,
-			 wnd->nmb->adapter_ctx);
+		LIN2WIN1(miniport->enable_interrupts, wnd->nmb->adapter_ctx);
 }
 
 static irqreturn_t ndis_isr(int irq, void *data, struct pt_regs *pt_regs)
@@ -1921,7 +1913,6 @@ STDCALL void WRAP_EXPORT(NdisMIndicateStatus)
 	case NDIS_STATUS_MEDIA_DISCONNECT:
 		if (wnd->link_status != 0) {
 			wnd->link_status = 0;
-			wnd->tx_ok = 0;
 			set_bit(LINK_STATUS_CHANGED, &wnd->wrap_ndis_work);
 			schedule_wrap_work(&wnd->wrap_ndis_worker);
 		}
@@ -1929,7 +1920,6 @@ STDCALL void WRAP_EXPORT(NdisMIndicateStatus)
 	case NDIS_STATUS_MEDIA_CONNECT:
 		if (wnd->link_status != 1) {
 			wnd->link_status = 1;
-			wnd->tx_ok = 1;
 			set_bit(LINK_STATUS_CHANGED, &wnd->wrap_ndis_work);
 			schedule_wrap_work(&wnd->wrap_ndis_worker);
 		}
@@ -1982,6 +1972,10 @@ STDCALL void WRAP_EXPORT(NdisMIndicateStatus)
 				  "num_cand %ld",
 				  cand->version, cand->num_candidates);
 			for (i = 0; i < cand->num_candidates; i++) {
+#if WIRELESS_EXT > 17
+				struct iw_pmkid_cand pcand;
+				union iwreq_data wrqu;
+#endif
 				struct ndis_pmkid_candidate *c =
 					&cand->candidates[i];
 				if ((u8 *)(c + 1) > end) {
@@ -1992,24 +1986,16 @@ STDCALL void WRAP_EXPORT(NdisMIndicateStatus)
 				DBGTRACE2("%ld: " MACSTR " 0x%lx",
 					  i, MAC2STR(c->bssid), c->flags);
 #if WIRELESS_EXT > 17
-				{
-					struct iw_pmkid_cand pcand;
-					union iwreq_data wrqu;
-					memset(&pcand, 0, sizeof(pcand));
-					if (c->flags & 0x01)
-						pcand.flags |=
-							IW_PMKID_CAND_PREAUTH;
-					pcand.index = i;
-					memcpy(pcand.bssid.sa_data, c->bssid,
-					       ETH_ALEN);
+				memset(&pcand, 0, sizeof(pcand));
+				if (c->flags & 0x01)
+					pcand.flags |= IW_PMKID_CAND_PREAUTH;
+				pcand.index = i;
+				memcpy(pcand.bssid.sa_data, c->bssid, ETH_ALEN);
 
-					memset(&wrqu, 0, sizeof(wrqu));
-					wrqu.data.length = sizeof(pcand);
-					wireless_send_event(wnd->net_dev,
-							    IWEVPMKIDCAND,
-							    &wrqu,
-							    (u8 *)&pcand);
-				}
+				memset(&wrqu, 0, sizeof(wrqu));
+				wrqu.data.length = sizeof(pcand);
+				wireless_send_event(wnd->net_dev, IWEVPMKIDCAND,
+						    &wrqu, (u8 *)&pcand);
 #endif
 			}
 			break;
@@ -2081,6 +2067,7 @@ NdisMIndicateReceivePacket(struct ndis_miniport_block *nmb,
 	int i, length, total_length;
 	struct ndis_packet_oob_data *oob_data;
 	void *virt;
+	struct ndis_tcp_ip_checksum_packet_info *rx_csum_info;
 
 	TRACEENTER3("%p, %d", nmb, nr_packets);
 	wnd = nmb->wnd;
@@ -2095,6 +2082,7 @@ NdisMIndicateReceivePacket(struct ndis_miniport_block *nmb,
 						 &length, &total_length,
 						 NormalPagePriority);
 		DBGTRACE3("%d, %d", length, total_length);
+		oob_data = NDIS_PACKET_OOB_DATA(packet);
 		skb = dev_alloc_skb(total_length);
 		if (skb) {
 			skb->dev = wnd->net_dev;
@@ -2107,16 +2095,24 @@ NdisMIndicateReceivePacket(struct ndis_miniport_block *nmb,
 			skb->protocol = eth_type_trans(skb, wnd->net_dev);
 			wnd->stats.rx_bytes += total_length;
 			wnd->stats.rx_packets++;
+			rx_csum_info =
+				oob_data->extension.info[TcpIpChecksumPacketInfo];
+			if (wnd->rx_csum.ip_csum && rx_csum_info &&
+			    (rx_csum_info->rx.tcp_succeeded ||
+			     rx_csum_info->rx.ip_succeeded ||
+			     rx_csum_info->rx.udp_succeeded)) {
+				skb->ip_summed = CHECKSUM_HW;
+				skb->csum = rx_csum_info->value;
+			}
 			netif_rx(skb);
 		} else {
 			WARNING("couldn't allocate skb; packet dropped");
 			wnd->stats.rx_dropped++;
 		}
 
-		oob_data = NDIS_PACKET_OOB_DATA(packet);
 		/* serialized drivers check the status upon return
 		 * from this function */
-		if (test_bit(ATTR_SERIALIZED, &wnd->attributes)) {
+		if (!(wnd->attributes & NDIS_ATTRIBUTE_DESERIALIZE)) {
 			oob_data->status = NDIS_STATUS_SUCCESS;
 			continue;
 		}
@@ -2136,7 +2132,8 @@ NdisMIndicateReceivePacket(struct ndis_miniport_block *nmb,
 		 * MiniportReturnPacket from here is not correct - the
 		 * driver doesn't expect it (at least Centrino driver
 		 * crashes) */
-		schedule_ntos_work_item(return_packet, wnd, packet, FALSE);
+		schedule_ntos_work_item(return_packet, wnd, packet,
+					WORKER_FUNC_LINUX);
 	}
 	TRACEEXIT3(return);
 }
@@ -2158,8 +2155,9 @@ NdisMSendComplete(struct ndis_miniport_block *nmb, struct ndis_packet *packet,
 	struct wrap_ndis_device *wnd = nmb->wnd;
 	TRACEENTER3("%p, %08x", packet, status);
 	free_tx_packet(wnd, packet, status);
-	/* In case a serialized driver has requested a pause by returning
-	 * NDIS_STATUS_RESOURCES we need to give the send-code a kick again.
+	/* In case a serialized driver has earlier requested a pause
+	 * by returning NDIS_STATUS_RESOURCES during
+	 * MiniportSend(Packets), wakeup tx worker now.
 	 */
 	if (wnd->tx_ok == 0) {
 		wnd->tx_ok = 1;
@@ -2201,6 +2199,7 @@ EthRxIndicateHandler(struct ndis_miniport_block *nmb, void *rx_ctx,
 	struct wrap_ndis_device *wnd;
 	unsigned int skb_size = 0;
 	KIRQL irql;
+	struct ndis_packet_oob_data *oob_data;
 
 	TRACEENTER3("nmb = %p, rx_ctx = %p, buf = %p, size = %d, "
 		    "buf = %p, size = %d, packet = %d",
@@ -2225,18 +2224,17 @@ EthRxIndicateHandler(struct ndis_miniport_block *nmb, void *rx_ctx,
 			wnd->stats.rx_dropped++;
 			TRACEEXIT3(return);
 		}
-
+		oob_data = NDIS_PACKET_OOB_DATA(packet);
 		miniport = &wnd->wd->driver->ndis_driver->miniport;
 		irql = raise_irql(DISPATCH_LEVEL);
-		res = LIN2WIN6(miniport->tx_data, packet, &bytes_txed,
-			       nmb, rx_ctx, look_ahead_size,
-			       packet_size);
+		res = LIN2WIN6(miniport->tx_data, packet, &bytes_txed, nmb,
+			       rx_ctx, look_ahead_size, packet_size);
 		lower_irql(irql);
 		DBGTRACE3("%d, %d, %d", header_size, look_ahead_size,
 			  bytes_txed);
 		if (res == NDIS_STATUS_SUCCESS) {
 			ndis_buffer *buffer;
-			skb = dev_alloc_skb(header_size+look_ahead_size+
+			skb = dev_alloc_skb(header_size + look_ahead_size +
 					    bytes_txed);
 			if (skb) {
 				memcpy(skb_put(skb, header_size), header,
@@ -2256,9 +2254,7 @@ EthRxIndicateHandler(struct ndis_miniport_block *nmb, void *rx_ctx,
 				NdisFreePacket(packet);
 			}
 		} else if (res == NDIS_STATUS_PENDING) {
-			struct ndis_packet_oob_data *oob_data;
 			/* driver will call td_complete */
-			oob_data = NDIS_PACKET_OOB_DATA(packet);
 			oob_data->look_ahead = kmalloc(look_ahead_size,
 						       GFP_ATOMIC);
 			if (!oob_data->look_ahead) {
@@ -2272,6 +2268,7 @@ EthRxIndicateHandler(struct ndis_miniport_block *nmb, void *rx_ctx,
 			memcpy(oob_data->look_ahead, look_ahead,
 			       look_ahead_size);
 			oob_data->look_ahead_size = look_ahead_size;
+			TRACEEXIT3(return);
 		} else {
 			WARNING("packet dropped: %08X", res);
 			NdisFreePacket(packet);
@@ -2288,7 +2285,7 @@ EthRxIndicateHandler(struct ndis_miniport_block *nmb, void *rx_ctx,
 		}
 	}
 
-	if (skb && skb_size > 0) {
+	if (skb) {
 		skb->dev = wnd->net_dev;
 		skb->protocol = eth_type_trans(skb, wnd->net_dev);
 		wnd->stats.rx_bytes += skb_size;
@@ -2313,6 +2310,7 @@ NdisMTransferDataComplete(struct ndis_miniport_block *nmb,
 	unsigned int skb_size;
 	struct ndis_packet_oob_data *oob_data;
 	ndis_buffer *buffer;
+	struct ndis_tcp_ip_checksum_packet_info *rx_csum_info;
 
 	TRACEENTER3("wnd = %p, packet = %p, bytes_txed = %d",
 		    wnd, packet, bytes_txed);
@@ -2336,8 +2334,8 @@ NdisMTransferDataComplete(struct ndis_miniport_block *nmb,
 	}
 
 	skb->dev = wnd->net_dev;
-	memcpy(skb_put(skb, sizeof(oob_data->header)),
-	       oob_data->header, sizeof(oob_data->header));
+	memcpy(skb_put(skb, sizeof(oob_data->header)), oob_data->header,
+	       sizeof(oob_data->header));
 	memcpy(skb_put(skb, oob_data->look_ahead_size), oob_data->look_ahead,
 	       oob_data->look_ahead_size);
 	buffer = packet->private.buffer_head;
@@ -2349,10 +2347,16 @@ NdisMTransferDataComplete(struct ndis_miniport_block *nmb,
 	}
 	kfree(oob_data->look_ahead);
 	NdisFreePacket(packet);
-	skb_put(skb, skb_size);
 	skb->protocol = eth_type_trans(skb, wnd->net_dev);
 	wnd->stats.rx_bytes += skb_size;
 	wnd->stats.rx_packets++;
+	rx_csum_info = oob_data->extension.info[TcpIpChecksumPacketInfo];
+	if (wnd->rx_csum.ip_csum && rx_csum_info &&
+	    (rx_csum_info->rx.tcp_succeeded || rx_csum_info->rx.ip_succeeded ||
+	     rx_csum_info->rx.udp_succeeded)) {
+		skb->ip_summed = CHECKSUM_HW;
+		skb->csum = rx_csum_info->value;
+	}
 	netif_rx(skb);
 }
 
@@ -2548,7 +2552,6 @@ NdisMResetComplete(struct ndis_miniport_block *nmb, NDIS_STATUS status,
 
 	TRACEENTER3("status: %08X, reset status: %u", status,
 		    address_reset);
-
 	wnd->ndis_comm_status = status;
 	wnd->ndis_comm_done = 1 + address_reset;
 	wake_up(&wnd->ndis_comm_wq);
@@ -2565,13 +2568,10 @@ STDCALL NDIS_STATUS WRAP_EXPORT(NdisScheduleWorkItem)
 	ndis_work_entry = kmalloc(sizeof(*ndis_work_entry), GFP_ATOMIC);
 	if (!ndis_work_entry)
 		BUG();
-
 	ndis_work_entry->ndis_work_item = ndis_work_item;
-
 	irql = nt_spin_lock_irql(&ndis_work_list_lock, DISPATCH_LEVEL);
 	InsertTailList(&ndis_worker_list, &ndis_work_entry->list);
 	nt_spin_unlock_irql(&ndis_work_list_lock, irql);
-
 	WORKTRACE("scheduling %p", ndis_work_item);
 	schedule_ndis_work(&ndis_work);
 	TRACEEXIT3(return NDIS_STATUS_SUCCESS);
