@@ -96,6 +96,7 @@ static void update_user_shared_data_proc(unsigned long data);
 static int add_bus_driver(const char *name);
 static void free_all_objects(void);
 static BOOLEAN queue_kdpc(struct kdpc *kdpc);
+BOOLEAN dequeue_kdpc(struct kdpc *kdpc);
 
 WRAP_EXPORT_MAP("KeTickCount", &jiffies);
 
@@ -105,21 +106,6 @@ WRAP_EXPORT_MAP("NlsMbCodePageTag", FALSE);
 struct workqueue_struct *ntos_wq;
 #endif
 
-/* Atheros driver allocates 200336 bytes in atomic context when
- * initializing. To satisfy this request, we allocate 200KB block
- * before driver initialization function is called and give this block
- * to driver instead of allocating. */
-
-#define ATHEROS_VMALLOC_BLOCK_SIZE (200 * 1024)
-struct vmalloc_block {
-	void *addr;
-	BOOLEAN in_use;
-	struct vmalloc_block *next;
-};
-static struct vmalloc_block *vmalloc_blocks;
-
-static STDCALL void wrap_vfree(void *addr, void *ctx);
-
 int ntoskernel_init(void)
 {
 	struct timeval now;
@@ -173,7 +159,6 @@ int ntoskernel_init(void)
 	init_timer(&shared_data_timer);
 	shared_data_timer.function = &update_user_shared_data_proc;
 #endif
-	vmalloc_blocks = NULL;
 	return 0;
 }
 
@@ -188,30 +173,6 @@ int ntoskernel_init_device(struct wrap_device *wd)
 	 * timer */
 	mod_timer(&shared_data_timer, jiffies + 1);
 #endif
-	/* if it is Atheros, allocate 200KB block and put it on a list
-	 * so this block can be given to driver when it requires it in
-	 * atomic context */
-	if (wd->vendor == 0x168c) {
-		struct vmalloc_block *vm_block;
-		vm_block = kmalloc(sizeof(*vm_block), GFP_KERNEL);
-		if (!vm_block)
-			WARNING("couldn't allocate memory");
-		else {
-			vm_block->addr = vmalloc(ATHEROS_VMALLOC_BLOCK_SIZE);
-			if (!vm_block->addr) {
-				WARNING("couldn't allocate memory");
-				kfree(vm_block);
-			} else {
-				KIRQL irql;
-				vm_block->in_use = FALSE;
-				irql = nt_spin_lock_irql(&ntoskernel_lock,
-							 DISPATCH_LEVEL);
-				vm_block->next = vmalloc_blocks;
-				vmalloc_blocks = vm_block;
-				nt_spin_unlock_irql(&ntoskernel_lock, irql);
-			}
-		}
-	}
 	return 0;
 }
 
@@ -241,31 +202,7 @@ void ntoskernel_exit_device(struct wrap_device *wd)
 			WARNING("Buggy Windows driver left timer %p running",
 				&wrap_timer->timer);
 		memset(wrap_timer, 0, sizeof(*wrap_timer));
-		wrap_kfree(wrap_timer);
-	}
-	if (wd->vendor == 0x168c) {
-		struct vmalloc_block *cur_vm_block, *prev_vm_block;
-		/* release one vm block that is free - since all
-		 * blocks are of same size, it doesn't matter which */
-		irql = nt_spin_lock_irql(&ntoskernel_lock, DISPATCH_LEVEL);
-		prev_vm_block = cur_vm_block = vmalloc_blocks;
-		while (cur_vm_block) {
-			if (cur_vm_block->in_use == FALSE) {
-				if (cur_vm_block == vmalloc_blocks)
-					vmalloc_blocks = vmalloc_blocks->next;
-				else
-					prev_vm_block->next =
-						cur_vm_block->next;
-				schedule_ntos_work_item(wrap_vfree,
-							cur_vm_block->addr,
-							NULL, FALSE);
-				kfree(cur_vm_block);
-				break;
-			}
-			prev_vm_block = cur_vm_block;
-			cur_vm_block = cur_vm_block->next;
-		}
-		nt_spin_unlock_irql(&ntoskernel_lock, irql);
+		slack_kfree(wrap_timer);
 	}
 	TRACEEXIT2(return);
 }
@@ -291,7 +228,7 @@ void ntoskernel_exit(void)
 			WARNING("Buggy Windows driver left timer %p running",
 				&wrap_timer->timer);
 		memset(wrap_timer, 0, sizeof(*wrap_timer));
-		wrap_kfree(wrap_timer);
+		slack_kfree(wrap_timer);
 	}
 
 	DBGTRACE2("freeing MDLs");
@@ -350,8 +287,6 @@ void ntoskernel_exit(void)
 	if (ntos_wq)
 		destroy_workqueue(ntos_wq);
 #endif
-	if (vmalloc_blocks)
-		ERROR("vmalloc blocks leaking!");
 	TRACEEXIT2(return);
 }
 
@@ -755,19 +690,19 @@ static void timer_proc(unsigned long data)
 	BUG_ON(wrap_timer->wrap_timer_magic != WRAP_TIMER_MAGIC);
 	BUG_ON(nt_timer->wrap_timer_magic != WRAP_TIMER_MAGIC);
 #endif
-	kdpc = nt_timer->kdpc;
 	KeSetEvent((struct nt_event *)nt_timer, 0, FALSE);
-	if (kdpc && kdpc->func) {
-		DBGTRACE5("kdpc %p (%p)", kdpc, kdpc->func);
-		LIN2WIN4(kdpc->func, kdpc, kdpc->ctx,
-			 kdpc->arg1, kdpc->arg2);
-	}
+
+	kdpc = nt_timer->kdpc;
+	if (kdpc && kdpc->func)
+//		queue_kdpc(kdpc);
+		LIN2WIN4(kdpc->func, kdpc, kdpc->ctx, kdpc->arg1, kdpc->arg2);
 
 	nt_spin_lock(&timer_lock);
 	if (wrap_timer->repeat)
 		mod_timer(&wrap_timer->timer, jiffies + wrap_timer->repeat);
 	nt_spin_unlock(&timer_lock);
 
+
 	TRACEEXIT5(return);
 }
 
@@ -783,11 +718,11 @@ void wrap_init_timer(struct nt_timer *nt_timer, enum timer_type type,
 	 * set and not allocate, but it is not guaranteed always to be
 	 * safe */
 	TRACEENTER5("%p", nt_timer);
-	/* we allocate memory for wrap_timer behind driver's back
-	 * and there is no NDIS/DDK function where this memory can be
-	 * freed, so we use wrap_kmalloc so it gets freed when driver
+	/* we allocate memory for wrap_timer behind driver's back and
+	 * there is no NDIS/DDK function where this memory can be
+	 * freed, so we use slack_kmalloc so it gets freed when driver
 	 * is unloaded */
-	wrap_timer = wrap_kmalloc(sizeof(*wrap_timer));
+	wrap_timer = slack_kmalloc(sizeof(*wrap_timer));
 	if (!wrap_timer) {
 		ERROR("couldn't allocate memory for timer");
 		return;
@@ -927,7 +862,7 @@ STDCALL void WRAP_EXPORT(KeInitializeDpc)
 {
 	TRACEENTER3("%p, %p, %p", kdpc, func, ctx);
 	memset(kdpc, 0, sizeof(*kdpc));
-	kdpc->number = 0;
+	kdpc->nr_cpu = FALSE;
 	kdpc->func = func;
 	kdpc->ctx  = ctx;
 	InitializeListHead(&kdpc->list);
@@ -947,7 +882,7 @@ static void kdpc_worker(void *data)
 			break;
 		}
 		kdpc = container_of(entry, struct kdpc, list);
-		kdpc->number = 0;
+		kdpc->nr_cpu = FALSE;
 		nt_spin_unlock_irql(&kdpc_list_lock, irql);
 		DBGTRACE5("%p, %p, %p, %p, %p", kdpc, kdpc->func, kdpc->ctx,
 			  kdpc->arg1, kdpc->arg2);
@@ -971,18 +906,16 @@ static BOOLEAN queue_kdpc(struct kdpc *kdpc)
 	if (!kdpc)
 		return FALSE;
 	irql = nt_spin_lock_irql(&kdpc_list_lock, DISPATCH_LEVEL);
-	if (kdpc->number) {
-		if (kdpc->number != 1)
-			ERROR("kdpc->number: %d", kdpc->number);
+	if (kdpc->nr_cpu) {
+		if (kdpc->nr_cpu != TRUE)
+			ERROR("kdpc->nr_cpu: %d", kdpc->nr_cpu);
 		ret = FALSE;
 	} else {
-		kdpc->number = 1;
+		ret = kdpc->nr_cpu = TRUE;
 		InsertTailList(&kdpc_list, &kdpc->list);
-		ret = TRUE;
+		schedule_ntos_work(&kdpc_work);
 	}
 	nt_spin_unlock_irql(&kdpc_list_lock, irql);
-	if (ret == TRUE)
-		schedule_ntos_work(&kdpc_work);
 	TRACEEXIT5(return ret);
 }
 
@@ -991,14 +924,15 @@ BOOLEAN dequeue_kdpc(struct kdpc *kdpc)
 	BOOLEAN ret;
 	KIRQL irql;
 
+	TRACEENTER5("%p", kdpc);
 	if (!kdpc)
 		return FALSE;
 	irql = nt_spin_lock_irql(&kdpc_list_lock, DISPATCH_LEVEL);
-	if (kdpc->number) {
-		if (kdpc->number != 1)
-			ERROR("kdpc->number: %d", kdpc->number);
+	if (kdpc->nr_cpu) {
+		if (kdpc->nr_cpu != TRUE)
+			ERROR("kdpc->nr_cpu: %d", kdpc->nr_cpu);
+		kdpc->nr_cpu = FALSE;
 		RemoveEntryList(&kdpc->list);
-		kdpc->number = 0;
 		ret = TRUE;
 	} else
 		ret = FALSE;
@@ -1047,7 +981,7 @@ static void ntos_work_item_worker(void *data)
 		func = ntos_work_item->func;
 		WORKTRACE("%p: executing %p, %p, %p", current, func,
 			  ntos_work_item->arg1, ntos_work_item->arg2);
-		if (ntos_work_item->win_func == TRUE)
+		if (ntos_work_item->func_type == WORKER_FUNC_WIN)
 			LIN2WIN2(func, ntos_work_item->arg1,
 				 ntos_work_item->arg2);
 		else
@@ -1058,7 +992,7 @@ static void ntos_work_item_worker(void *data)
 }
 
 int schedule_ntos_work_item(NTOS_WORK_FUNC func, void *arg1, void *arg2,
-			    BOOLEAN win_func)
+			    enum worker_func_type func_type)
 {
 	struct ntos_work_item *ntos_work_item;
 	KIRQL irql;
@@ -1072,7 +1006,7 @@ int schedule_ntos_work_item(NTOS_WORK_FUNC func, void *arg1, void *arg2,
 	ntos_work_item->func = func;
 	ntos_work_item->arg1 = arg1;
 	ntos_work_item->arg2 = arg2;
-	ntos_work_item->win_func = win_func;
+	ntos_work_item->func_type = func_type;
 	irql = nt_spin_lock_irql(&ntos_work_item_list_lock, DISPATCH_LEVEL);
 	InsertTailList(&ntos_work_item_list, &ntos_work_item->list);
 	nt_spin_unlock_irql(&ntos_work_item_list_lock, irql);
@@ -1142,6 +1076,8 @@ STDCALL KIRQL WRAP_EXPORT(KeAcquireSpinLockRaiseToDpc)
 	return nt_spin_lock_irql(lock, DISPATCH_LEVEL);
 }
 
+#undef ExAllocatePoolWithTag
+
 STDCALL void *WRAP_EXPORT(ExAllocatePoolWithTag)
 	(enum pool_type pool_type, SIZE_T size, ULONG tag)
 {
@@ -1156,36 +1092,17 @@ STDCALL void *WRAP_EXPORT(ExAllocatePoolWithTag)
 		else
 			addr = kmalloc(size, GFP_ATOMIC);
 	} else {
-		addr = NULL;
-		if (current_irql() > PASSIVE_LEVEL &&
-		    size <= ATHEROS_VMALLOC_BLOCK_SIZE) {
-			KIRQL irql;
-			struct vmalloc_block *vm_block;
-			irql = nt_spin_lock_irql(&ntoskernel_lock,
-						 DISPATCH_LEVEL);
-			vm_block = vmalloc_blocks;
-			while (vm_block) {
-				if (vm_block->in_use == FALSE) {
-					vm_block->in_use = TRUE;
-					addr = vm_block->addr;
-					break;
-				}
-				vm_block = vm_block->next;
-			}
-			nt_spin_unlock_irql(&ntoskernel_lock, irql);
-		}
-		if (!addr) {
-			if (current_irql() > PASSIVE_LEVEL)
-				WARNING("Windows driver allocating %lu bytes "
-					"in atomic context", size);
+		if (current_irql() < DISPATCH_LEVEL)
 			addr = vmalloc(size);
-		}
+		else
+			addr = __vmalloc(size, GFP_ATOMIC | __GFP_HIGHMEM,
+					 PAGE_KERNEL);
 	}
 	DBGTRACE4("addr: %p, %lu", addr, size);
 	TRACEEXIT4(return addr);
 }
 
-static STDCALL void wrap_vfree(void *addr, void *ctx)
+static STDCALL void vfree_nonatomic(void *addr, void *ctx)
 {
 	vfree(addr);
 }
@@ -1198,25 +1115,11 @@ STDCALL void WRAP_EXPORT(ExFreePool)
 	    (unsigned long)addr >= VMALLOC_END)
 		kfree(addr);
 	else {
-		KIRQL irql;
-		struct vmalloc_block *vm_block;
-		irql = nt_spin_lock_irql(&ntoskernel_lock, DISPATCH_LEVEL);
-		vm_block = vmalloc_blocks;
-		while (vm_block) {
-			if (vm_block->addr == addr) {
-				vm_block->in_use = FALSE;
-				break;
-			}
-			vm_block = vm_block->next;
-		}
-		nt_spin_unlock_irql(&ntoskernel_lock, irql);
-		if (!vm_block) {
-			if (in_interrupt())
-				schedule_ntos_work_item(wrap_vfree, addr,
-							NULL, FALSE);
-			else
-				vfree(addr);
-		}
+		if (in_interrupt())
+			schedule_ntos_work_item(vfree_nonatomic, addr,
+						NULL, FALSE);
+		else
+			vfree(addr);
 	}
 	return;
 }
@@ -1284,6 +1187,11 @@ STDCALL void WRAP_EXPORT(ExDeleteNPagedLookasideList)
 	TRACEEXIT3(return);
 }
 
+#ifdef ALLOC_DEBUG
+#define ExAllocatePoolWithTag(pool_type, size, tag)			\
+	wrap_ExAllocatePoolWithTag(pool_type, size, tag, __FILE__, __LINE__)
+#endif
+
 STDCALL NTSTATUS WRAP_EXPORT(ExCreateCallback)
 	(struct callback_object **object, struct object_attributes *attributes,
 	 BOOLEAN create, BOOLEAN allow_multiple_callbacks)
@@ -1424,13 +1332,13 @@ static void wakeup_threads(struct dispatcher_header *dh)
 		assert(wb->object == NULL);
 		if (wb->thread &&
 		    check_grab_signaled_state(dh, wb->thread, 1)) {
-			struct thread_event_wait *event_wait = wb->event_wait;
-			EVENTTRACE("%p: waking up task %p for %p", event_wait,
+			struct thread_event_wait *thread_wait = wb->thread_wait;
+			EVENTTRACE("%p: waking up task %p for %p", thread_wait,
 				   wb->thread, dh);
 			RemoveEntryList(&wb->list);
 			wb->object = dh;
-			event_wait->done = 1;
-			wake_up(&event_wait->wq_head);
+			thread_wait->done = 1;
+			wake_up(&thread_wait->wq_head);
 			if (dh->type == SynchronizationObject)
 				break;
 		} else
@@ -1442,7 +1350,7 @@ static void wakeup_threads(struct dispatcher_header *dh)
 /* We need workqueue to implement KeWaitFor routines
  * below. (get/put)_thread_event_wait give/take back a workqueue. Both
  * these are called holding dispatcher spinlock, so no locking here */
-static struct thread_event_wait *get_thread_event_wait(void)
+static inline struct thread_event_wait *get_thread_event_wait(void)
 {
 	struct thread_event_wait *thread_event_wait;
 
@@ -1493,7 +1401,7 @@ STDCALL NTSTATUS WRAP_EXPORT(KeWaitForMultipleObjects)
 	struct wait_block *wb, wb_array[THREAD_WAIT_OBJECTS];
 	struct dispatcher_header *dh;
 	struct task_struct *thread;
-	struct thread_event_wait *event_wait;
+	struct thread_event_wait *thread_wait;
 	KIRQL irql;
 
 	thread = current;
@@ -1547,13 +1455,13 @@ STDCALL NTSTATUS WRAP_EXPORT(KeWaitForMultipleObjects)
 			nt_spin_unlock_irql(&dispatcher_lock, irql);
 			EVENTEXIT(return STATUS_TIMEOUT);
 		}
-		event_wait = get_thread_event_wait();
-		if (!event_wait) {
+		thread_wait = get_thread_event_wait();
+		if (!thread_wait) {
 			nt_spin_unlock_irql(&dispatcher_lock, irql);
 			EVENTEXIT(return STATUS_RESOURCES);
 		}
 	} else
-		event_wait = NULL;
+		thread_wait = NULL;
 
 	/* get the list of objects the thread needs to wait on and add
 	 * the thread on the wait list for each such object */
@@ -1563,7 +1471,7 @@ STDCALL NTSTATUS WRAP_EXPORT(KeWaitForMultipleObjects)
 		EVENTTRACE("%p: event %p state: %d",
 			   thread, dh, dh->signal_state);
 		wb[i].object = NULL;
-		wb[i].event_wait = event_wait;
+		wb[i].thread_wait = thread_wait;
 		if (check_grab_signaled_state(dh, thread, 1)) {
 			EVENTTRACE("%p: event %p already signaled: %d",
 				   thread, dh, dh->signal_state);
@@ -1571,7 +1479,7 @@ STDCALL NTSTATUS WRAP_EXPORT(KeWaitForMultipleObjects)
 			wb[i].thread = NULL;
 		} else {
 			assert(timeout == NULL || *timeout != 0);
-			assert(event_wait != NULL);
+			assert(thread_wait != NULL);
 			wb[i].thread = thread;
 			EVENTTRACE("%p: need to wait on event %p", thread, dh);
 			InsertTailList(&dh->wait_blocks, &wb[i].list);
@@ -1579,10 +1487,10 @@ STDCALL NTSTATUS WRAP_EXPORT(KeWaitForMultipleObjects)
 	}
 	nt_spin_unlock_irql(&dispatcher_lock, irql);
 	if (wait_count == 0) {
-		assert(event_wait == NULL);
+		assert(thread_wait == NULL);
 		EVENTEXIT(return STATUS_SUCCESS);
 	}
-	assert(event_wait);
+	assert(thread_wait);
 
 	assert(timeout == NULL || *timeout != 0);
 	if (timeout == NULL)
@@ -1590,30 +1498,29 @@ STDCALL NTSTATUS WRAP_EXPORT(KeWaitForMultipleObjects)
 	else
 		wait_jiffies = SYSTEM_TIME_TO_HZ(*timeout) + 1;
 	EVENTTRACE("%p: sleeping for %ld on %p",
-		   thread, wait_jiffies, event_wait);
+		   thread, wait_jiffies, thread_wait);
 
 	while (wait_count) {
 		if (wait_jiffies) {
 			res = wait_event_interruptible_timeout(
-				event_wait->wq_head,
-				(event_wait->done == 1),
+				thread_wait->wq_head, (thread_wait->done == 1),
 				wait_jiffies);
 		} else {
 			wait_event_interruptible(
-				event_wait->wq_head,
-				(event_wait->done == 1));
+				thread_wait->wq_head,(thread_wait->done == 1));
 			/* mark that it didn't timeout */
 			res = 1;
 		}
+		thread_wait->done = 0;
 		irql = nt_spin_lock_irql(&dispatcher_lock, DISPATCH_LEVEL);
 		if (signal_pending(current))
 			res = -ERESTARTSYS;
 		EVENTTRACE("%p woke up on %p, res = %d, done: %d", thread,
-			   event_wait, res, event_wait->done);
+			   thread_wait, res, thread_wait->done);
 #ifdef EVENT_DEBUG
-		if (event_wait->task != current)
-			ERROR("%p: argh, task %p should be %p", event_wait,
-			      event_wait->task, current);
+		if (thread_wait->task != current)
+			ERROR("%p: argh, task %p should be %p", thread_wait,
+			      thread_wait->task, current);
 #endif
 //		assert(res < 0 && alertable);
 		if (res <= 0) {
@@ -1625,7 +1532,7 @@ STDCALL NTSTATUS WRAP_EXPORT(KeWaitForMultipleObjects)
 					   thread, object[i], wb[i].object);
 				RemoveEntryList(&wb[i].list);
 			}
-			put_thread_event_wait(event_wait);
+			put_thread_event_wait(thread_wait);
 			nt_spin_unlock_irql(&dispatcher_lock, irql);
 			if (res < 0)
 				EVENTEXIT(return STATUS_ALERTED);
@@ -1658,13 +1565,13 @@ STDCALL NTSTATUS WRAP_EXPORT(KeWaitForMultipleObjects)
 				for (j = i; j < count; j++)
 					if (wb[j].thread)
 						RemoveEntryList(&wb[j].list);
-				put_thread_event_wait(event_wait);
+				put_thread_event_wait(thread_wait);
 				nt_spin_unlock_irql(&dispatcher_lock, irql);
 				EVENTEXIT(return STATUS_WAIT_0 + i);
 			}
 		}
 		if (wait_count == 0) {
-			put_thread_event_wait(event_wait);
+			put_thread_event_wait(thread_wait);
 			nt_spin_unlock_irql(&dispatcher_lock, irql);
 			EVENTEXIT(return STATUS_SUCCESS);
 		}
@@ -1672,7 +1579,6 @@ STDCALL NTSTATUS WRAP_EXPORT(KeWaitForMultipleObjects)
 		 * let it wait for remaining time and those objects */
 		/* we already set res to 1 if timeout was NULL, so
 		 * reinitialize wait_jiffies accordingly */
-		event_wait->done = 0;
 		if (timeout)
 			wait_jiffies = res;
 		else
@@ -1719,14 +1625,6 @@ STDCALL LONG WRAP_EXPORT(KeSetEvent)
 	nt_event->dh.signal_state = 1;
 	if (old_state == 0)
 		wakeup_threads(&nt_event->dh);
-	/* Synchronization objects are auto-reset - they don't stay
-	 * signalled. The DDK says they are reset after waking up one
-	 * thread. However, it is not clear if they should be reset if
-	 * there are no threads waiting. Am1772 driver expects that
-	 * they are reset (and not stay signalled) if no thread is
-	 * waiting */
-	if (nt_event->dh.type == SynchronizationObject)
-		nt_event->dh.signal_state = 0;
 	nt_spin_unlock_irql(&dispatcher_lock, irql);
 	EVENTEXIT(return old_state);
 }
@@ -1743,16 +1641,7 @@ STDCALL LONG WRAP_EXPORT(KeResetEvent)
 	(struct nt_event *nt_event)
 {
 	LONG old_state;
-#if 1
-	KIRQL irql;
-	EVENTENTER("event = %p", nt_event);
-	irql = nt_spin_lock_irql(&dispatcher_lock, DISPATCH_LEVEL);
-	old_state = nt_event->dh.signal_state;
-	nt_event->dh.signal_state = 0;
-	nt_spin_unlock_irql(&dispatcher_lock, irql);
-#else
 	old_state = xchg(&nt_event->dh.signal_state, 0);
-#endif
 	EVENTTRACE("old state: %d", old_state);
 	EVENTEXIT(return old_state);
 }
@@ -1966,6 +1855,9 @@ static int thread_trampoline(void *data)
 	thread = ctx.thread;
 	thread->task = current;
 	thread->pid = thread->task->pid;
+#ifdef PF_NOFREEZE
+	current->flags |= PF_NOFREEZE;
+#endif
 
 	DBGTRACE2("thread: %p, task: %p (%d)", thread, thread->task,
 		  thread->pid);
@@ -2247,7 +2139,7 @@ struct mdl *allocate_init_mdl(void *virt, ULONG length)
 		InsertHeadList(&wrap_mdl_list, &wrap_mdl->list);
 		nt_spin_unlock_irql(&ntoskernel_lock, irql);
 		mdl = (struct mdl *)wrap_mdl->mdl;
-		DBGTRACE3("allocated mdl cache: %p(%p), %p(%d)",
+		DBGTRACE3("allocated mdl from cache: %p(%p), %p(%d)",
 			  wrap_mdl, mdl, virt, length);
 		memset(mdl, 0, CACHE_MDL_SIZE);
 		MmInitializeMdl(mdl, virt, length);
@@ -2261,7 +2153,7 @@ struct mdl *allocate_init_mdl(void *virt, ULONG length)
 		if (!wrap_mdl)
 			return NULL;
 		mdl = (struct mdl *)wrap_mdl->mdl;
-		DBGTRACE3("allocated mdl cache: %p(%p), %p(%d)",
+		DBGTRACE3("allocated mdl from memory: %p(%p), %p(%d)",
 			  wrap_mdl, mdl, virt, length);
 		irql = nt_spin_lock_irql(&ntoskernel_lock, DISPATCH_LEVEL);
 		InsertHeadList(&wrap_mdl_list, &wrap_mdl->list);
@@ -2310,6 +2202,7 @@ STDCALL void WRAP_EXPORT(IoBuildPartialMdl)
 	(struct mdl *source, struct mdl *target, void *virt, ULONG length)
 {
 	MmInitializeMdl(target, virt, length);
+	target->flags |= MDL_PARTIAL;
 }
 
 /* FIXME: We don't update MDL to physical page mapping, since in Linux
@@ -2318,35 +2211,33 @@ STDCALL void WRAP_EXPORT(IoBuildPartialMdl)
 STDCALL void WRAP_EXPORT(MmBuildMdlForNonPagedPool)
 	(struct mdl *mdl)
 {
-	TRACEENTER4("%p", mdl);
-	mdl->flags |= MDL_SOURCE_IS_NONPAGED_POOL;
-	MmGetSystemAddressForMdl(mdl) = MmGetMdlVirtualAddress(mdl);
-	DBGTRACE4("%p, %p, %p, %d, %d", mdl, mdl->mappedsystemva, mdl->startva,
-		  mdl->byteoffset, mdl->bytecount);
-#if 0
-	{
 	PFN_NUMBER *mdl_pages;
 	int i, n;
-	void *start;
-
+	TRACEENTER4("%p", mdl);
+	/* already mapped */
+//	mdl->mappedsystemva = MmGetMdlVirtualAddress(mdl);
+	mdl->flags |= MDL_SOURCE_IS_NONPAGED_POOL;
+	DBGTRACE4("%p, %p, %p, %d, %d", mdl, mdl->mappedsystemva, mdl->startva,
+		  mdl->byteoffset, mdl->bytecount);
 	n = SPAN_PAGES(MmGetSystemAddressForMdl(mdl), MmGetMdlByteCount(mdl));
 	if (n > CACHE_MDL_PAGES)
 		WARNING("%p, %d, %d", MmGetSystemAddressForMdl(mdl),
 			MmGetMdlByteCount(mdl), n);
 	mdl_pages = MmGetMdlPfnArray(mdl);
-	start = MmGetSystemAddressForMdl(mdl);
 	for (i = 0; i < n; i++)
-		mdl_pages[i] =
-			page_to_pfn(virt_to_page(start + i * PAGE_SIZE));
-	}
-#endif
+		mdl_pages[i] = (ULONG_PTR)mdl->startva + (i * PAGE_SIZE);
 	return;
 }
 
 STDCALL void *WRAP_EXPORT(MmMapLockedPages)
 	(struct mdl *mdl, KPROCESSOR_MODE access_mode)
 {
+	/* already mapped */
+//	mdl->mappedsystemva = MmGetMdlVirtualAddress(mdl);
 	mdl->flags |= MDL_MAPPED_TO_SYSTEM_VA;
+	/* what is the need for MDL_PARTIAL_HAS_BEEN_MAPPED? */
+	if (mdl->flags & MDL_PARTIAL)
+		mdl->flags |= MDL_PARTIAL_HAS_BEEN_MAPPED;
 	return mdl->mappedsystemva;
 }
 
@@ -2369,6 +2260,7 @@ STDCALL void WRAP_EXPORT(MmProbeAndLockPages)
 	(struct mdl *mdl, KPROCESSOR_MODE access_mode,
 	 enum lock_operation operation)
 {
+	/* already locked */
 	mdl->flags |= MDL_PAGES_LOCKED;
 	return;
 }
@@ -2670,8 +2562,7 @@ STDCALL NTSTATUS WRAP_EXPORT(ZwQueryValueKey)
 	 ULONG length, ULONG *res_length)
 {
 	struct ansi_string ansi;
-	if (RtlUnicodeStringToAnsiString(&ansi, name, TRUE) ==
-	    STATUS_SUCCESS) {
+	if (RtlUnicodeStringToAnsiString(&ansi, name, TRUE) == STATUS_SUCCESS) {
 		DBGTRACE1("key: %s", ansi.buf);
 		RtlFreeAnsiString(&ansi);
 	}
