@@ -66,6 +66,8 @@
 #define addr_offset(drvr) (__builtin_return_address(0) -	\
 			   (drvr)->drv_obj->driver_start)
 
+#include "wrapmem.h"
+
 /* Workqueue / task queue backwards compatibility stuff */
 #if LINUX_VERSION_CODE > KERNEL_VERSION(2,5,41)
 #include <linux/workqueue.h>
@@ -392,10 +394,6 @@ do {									\
 #define irqreturn_t void
 #endif
 
-#ifndef free_netdev
-#define free_netdev kfree
-#endif
-
 #if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,9)
 #define WRAP_MODULE_PARM_INT(name, perm) module_param(name, int, perm)
 #define WRAP_MODULE_PARM_STRING(name, perm) module_param(name, charp, perm)
@@ -430,7 +428,7 @@ do {									\
 #define rt_task(p) ((p)->prio < MAX_RT_PRIO)
 #endif
 
-#define KMALLOC_THRESHOLD 131072
+#define KMALLOC_THRESHOLD 130000
 
 /* TICK is 100ns */
 #define TICKSPERSEC		10000000LL
@@ -485,11 +483,6 @@ struct wrap_export {
 #define POOL_TAG(A, B, C, D)					\
 	((ULONG)((A) + ((B) << 8) + ((C) << 16) + ((D) << 24)))
 
-struct wrap_alloc {
-	struct nt_list list;
-	void *ptr;
-};
-
 struct pe_image {
 	char name[MAX_DRIVER_NAME_LEN];
 	UINT (*entry)(struct driver_object *, struct unicode_string *) STDCALL;
@@ -516,12 +509,14 @@ struct wrap_timer {
 #endif
 };
 
+enum worker_func_type { WORKER_FUNC_LINUX, WORKER_FUNC_WIN };
+
 struct ntos_work_item {
 	struct nt_list list;
 	void *arg1;
 	void *arg2;
 	void (*func)(void *arg1, void *arg2) STDCALL;
-	BOOLEAN win_func;
+	enum worker_func_type func_type;
 };
 
 struct wrap_device_setting {
@@ -593,6 +588,7 @@ struct wrap_device {
 	struct nt_list timer_list;
 	NT_SPIN_LOCK timer_lock;
 	struct cm_resource_list *resource_list;
+	BOOLEAN surprise_removed;
 };
 
 /* Some drivers use worker entries to complete functions called from
@@ -650,8 +646,14 @@ struct driver_object *find_bus_driver(const char *name);
 STDCALL void WRITE_PORT_UCHAR(ULONG_PTR port, UCHAR value);
 STDCALL UCHAR READ_PORT_UCHAR(ULONG_PTR port);
 
+#undef ExAllocatePoolWithTag
 STDCALL void *ExAllocatePoolWithTag(enum pool_type pool_type, SIZE_T size,
 				    ULONG tag);
+#ifdef ALLOC_DEBUG
+#define ExAllocatePoolWithTag(pool_type, size, tag)			\
+	wrap_ExAllocatePoolWithTag(pool_type, size, tag, __FILE__, __LINE__)
+#endif
+
 STDCALL void ExFreePool(void *p);
 STDCALL ULONG MmSizeOfMdl(void *base, ULONG length);
 STDCALL void *MmMapIoSpace(PHYSICAL_ADDRESS phys_addr, SIZE_T size,
@@ -704,9 +706,12 @@ STDCALL NTSTATUS IoCreateSymbolicLink(struct unicode_string *link,
 STDCALL void IoDeleteDevice(struct device_object *dev);
 STDCALL void IoDetachDevice(struct device_object *topdev);
 STDCALL struct device_object *IoGetAttachedDevice(struct device_object *dev);
+STDCALL struct device_object *
+	IoGetAttachedDeviceReference(struct device_object *dev);
 STDCALL NTSTATUS
-IoAllocateDriverObjectExtension(struct driver_object *drv_obj,
-				void *client_id, ULONG extlen, void **ext);
+	IoAllocateDriverObjectExtension(struct driver_object *drv_obj,
+					void *client_id, ULONG extlen,
+					void **ext);
 STDCALL void *IoGetDriverObjectExtension(struct driver_object *drv,
 					 void *client_id);
 STDCALL struct device_object *IoAttachDeviceToDeviceStack
@@ -720,21 +725,35 @@ STDCALL void IoFreeIrp(struct irp *irp);
 STDCALL BOOLEAN IoCancelIrp(struct irp *irp);
 _FASTCALL NTSTATUS IofCallDriver
 	(FASTCALL_DECL_2(struct device_object *dev_obj, struct irp *irp));
-STDCALL struct irp *WRAP_EXPORT(IoBuildSynchronousFsdRequest)
-	(ULONG major_func, struct device_object *dev_obj, void *buf,
-	 ULONG length, LARGE_INTEGER *offset, struct nt_event *event,
-	 struct io_status_block *status);
-STDCALL struct irp *WRAP_EXPORT(IoBuildAsynchronousFsdRequest)
-	(ULONG major_func, struct device_object *dev_obj, void *buf,
-	 ULONG length, LARGE_INTEGER *offset,
-	 struct io_status_block *status);
+STDCALL struct irp *
+	IoBuildSynchronousFsdRequest(ULONG major_func,
+				     struct device_object *dev_obj, void *buf,
+				     ULONG length, LARGE_INTEGER *offset,
+				     struct nt_event *event,
+				     struct io_status_block *status);
+STDCALL struct irp *
+	IoBuildAsynchronousFsdRequest(ULONG major_func,
+				      struct device_object *dev_obj, void *buf,
+				      ULONG length, LARGE_INTEGER *offset,
+				      struct io_status_block *status);
 STDCALL NTSTATUS PoCallDriver(struct device_object *dev_obj, struct irp *irp);
 
+driver_dispatch_t IopInvalidDeviceRequest;
+
+STDCALL NTSTATUS IoPassIrpDown(struct device_object *dev_obj,
+			       struct irp *irp);
+STDCALL NTSTATUS IoSyncForwardIrp(struct device_object *dev_obj,
+				  struct irp *irp);
+STDCALL NTSTATUS IoAsyncForwardIrp(struct device_object *dev_obj,
+				   struct irp *irp);
+STDCALL NTSTATUS IoInvalidDeviceRequest(struct device_object *dev_obj,
+					struct irp *irp);
+
 struct nt_thread *get_current_nt_thread(void);
 u64 ticks_1601(void);
 
 int schedule_ntos_work_item(NTOS_WORK_FUNC func, void *arg1, void *arg2,
-			    BOOLEAN win_func);
+			    enum worker_func_type func_type);
 
 STDCALL KIRQL KeGetCurrentIrql(void);
 STDCALL void KeInitializeSpinLock(NT_SPIN_LOCK *lock);
@@ -774,8 +793,6 @@ STDCALL void RtlCopyUnicodeString
 	(struct unicode_string *dst, struct unicode_string *src);
 NOREGPARM SIZE_T _win_wcslen(const wchar_t *s);
 
-void *wrap_kmalloc(size_t size);
-void wrap_kfree(void *ptr);
 void wrap_init_timer(struct nt_timer *nt_timer, enum timer_type type,
 		     struct wrap_device *wd);
 BOOLEAN wrap_set_timer(struct nt_timer *nt_timer, unsigned long expires_hz,
@@ -832,9 +849,9 @@ _FASTCALL void ObfDereferenceObject(FASTCALL_DECL_1(void *object));
 void adjust_user_shared_data_addr(char *driver, unsigned long length);
 
 #define IoCompleteRequest(irp, prio)			\
-	IofCompleteRequest(FASTCALL_ARGS_2(irp, prio));
+	IofCompleteRequest(FASTCALL_ARGS_2(irp, prio))
 #define IoCallDriver(dev, irp)				\
-	IofCallDriver(FASTCALL_ARGS_2(dev, irp));
+	IofCallDriver(FASTCALL_ARGS_2(dev, irp))
 
 static inline KIRQL current_irql(void)
 {
@@ -986,8 +1003,8 @@ static inline ULONG SPAN_PAGES(void *ptr, SIZE_T length)
 {
 	ULONG n;
 
-	n = ((((ULONG_PTR)ptr) & (PAGE_SIZE - 1)) +
-	     length + (PAGE_SIZE - 1)) >> PAGE_SHIFT;
+	n = ((((ULONG_PTR)ptr) & (PAGE_SIZE - 1)) + length + (PAGE_SIZE - 1))
+		>> PAGE_SHIFT;
 
 	return n;
 }
