@@ -22,51 +22,32 @@ extern NT_SPIN_LOCK loader_lock;
 extern struct nt_list ndis_drivers;
 extern struct wrap_device *wrap_devices;
 
-/* IRP functions are called as Windows functions. For 64-bit, this
- * means arguments are in rcx, rdx etc., but Linux functions expect
- * them in rdi, rsi etc. So we need to put arguments back correctly
- * before touching arguments. We also assume that all arguments are
- * pointers (or register width). */
-
-/* called as Windows function, so call WIN2LIN3 before accessing
- * arguments */
-STDCALL NTSTATUS IrpStopCompletion(struct device_object *dev_obj,
-				   struct irp *irp, void *context)
+STDCALL NTSTATUS IoSendIrpTopDev(struct device_object *dev_obj, ULONG major_fn,
+				 ULONG minor_fn, struct io_stack_location *sl)
 {
-	WIN2LIN3(dev_obj, irp, context);
-
-	IOENTER("dev_obj: %p, irp: %p, context: %p", dev_obj, irp, context);
-	IOEXIT(return STATUS_MORE_PROCESSING_REQUIRED);
-}
-
-/* called as Windows function, so call WIN2LIN2 before accessing
- * arguments */
-STDCALL NTSTATUS IopPassIrpDown(struct device_object *dev_obj,
-				struct irp *irp)
-{
-	WIN2LIN2(dev_obj, irp);
-
-	IoSkipCurrentIrpStackLocation(irp);
-	return IoCallDriver(dev_obj, irp);
-}
-
-/* called as Windows function, so call WIN2LIN2 before accessing
- * arguments */
-STDCALL NTSTATUS IopInvalidDeviceRequest(struct device_object *dev_obj,
-					 struct irp *irp)
-{
-	struct io_stack_location *irp_sl;
 	NTSTATUS status;
+	struct nt_event event;
+	struct irp *irp;
+	struct io_stack_location *irp_sl;
+	struct device_object *top_dev = IoGetAttachedDeviceReference(dev_obj);
 
-	WIN2LIN2(dev_obj, irp);
-
-	irp_sl = IoGetCurrentIrpStackLocation(irp);
-	WARNING("IRP %d:%d not implemented",
-		irp_sl->major_fn, irp_sl->minor_fn);
-	irp->io_status.status = STATUS_SUCCESS;
+	KeInitializeEvent(&event, NotificationEvent, FALSE);
+	irp = IoBuildSynchronousFsdRequest(IRP_MJ_PNP, top_dev, NULL, 0, NULL,
+					   &event, NULL);
+	irp->io_status.status = STATUS_NOT_IMPLEMENTED;
 	irp->io_status.info = 0;
-	status = irp->io_status.status;
-	IoCompleteRequest(irp, IO_NO_INCREMENT);
+	irp_sl = IoGetNextIrpStackLocation(irp);
+	if (sl)
+		memcpy(irp_sl, sl, sizeof(*irp_sl));
+	irp_sl->major_fn = major_fn;
+	irp_sl->minor_fn = minor_fn;
+	status = IoCallDriver(top_dev, irp);
+	if (status == STATUS_PENDING) {
+		KeWaitForSingleObject(&event, Executive, KernelMode,
+				      FALSE, NULL);
+		status = irp->io_status.status;
+	}
+	ObDereferenceObject(top_dev);
 	return status;
 }
 
@@ -124,6 +105,9 @@ STDCALL NTSTATUS pdoDispatchPnp(struct device_object *pdo,
 	DBGTRACE2("fn %d:%d, wd: %p", irp_sl->major_fn, irp_sl->minor_fn, wd);
 	switch (irp_sl->minor_fn) {
 	case IRP_MN_START_DEVICE:
+	case IRP_MN_QUERY_STOP_DEVICE:
+	case IRP_MN_STOP_DEVICE:
+	case IRP_MN_QUERY_REMOVE_DEVICE:
 		status = STATUS_SUCCESS;
 		break;
 	case IRP_MN_QUERY_INTERFACE:
@@ -158,9 +142,6 @@ STDCALL NTSTATUS pdoDispatchPnp(struct device_object *pdo,
 		status = STATUS_NOT_IMPLEMENTED;
 #endif
 		break;
-	case IRP_MN_QUERY_REMOVE_DEVICE:
-		status = STATUS_SUCCESS;
-		break;
 	case IRP_MN_REMOVE_DEVICE:
 		ntoskernel_exit_device(wd);
 		if (wrap_is_pci_bus(wd->dev_bus_type)) {
@@ -177,7 +158,6 @@ STDCALL NTSTATUS pdoDispatchPnp(struct device_object *pdo,
 		if (wd->resource_list)
 			kfree(wd->resource_list);
 		wd->resource_list = NULL;
-		IoDeleteDevice(pdo);
 		status = STATUS_SUCCESS;
 		break;
 	default:
@@ -229,7 +209,9 @@ STDCALL NTSTATUS pdoDispatchPower(struct device_object *pdo,
 				pci_restore_state(pdev, wd->pci.pci_state);
 #endif
 			} else { // usb device
+#ifdef CONFIG_USB
 				wrap_resume_urbs(wd);
+#endif
 			}
 		} else {
 			DBGTRACE2("suspending device %p", wd);
@@ -241,7 +223,9 @@ STDCALL NTSTATUS pdoDispatchPower(struct device_object *pdo,
 				pci_save_state(pdev, wd->pci.pci_state);
 #endif
 			} else { // usb device
+#ifdef CONFIG_USB
 				wrap_suspend_urbs(wd);
+#endif
 			}
 		}
 		status = STATUS_SUCCESS;
@@ -272,7 +256,7 @@ static struct device_object *alloc_pdo(struct driver_object *drv_obj)
 	if (status != STATUS_SUCCESS)
 		return NULL;
 	for (i = 0; i <= IRP_MJ_MAXIMUM_FUNCTION; i++)
-		drv_obj->major_func[i] = IopInvalidDeviceRequest;
+		drv_obj->major_func[i] = IoInvalidDeviceRequest;
 	drv_obj->major_func[IRP_MJ_INTERNAL_DEVICE_CONTROL] =
 		pdoDispatchDeviceControl;
 	drv_obj->major_func[IRP_MJ_DEVICE_CONTROL] =
@@ -292,6 +276,7 @@ static int start_pdo(struct device_object *pdo)
 
 	TRACEENTER1("%p, %p", pdo, pdo->reserved);
 	wd = pdo->reserved;
+	wd->surprise_removed = TRUE;
 	if (ntoskernel_init_device(wd))
 		TRACEEXIT1(return -EINVAL);
 	DBGTRACE1("%d, %d", WRAP_BUS_TYPE(wd->dev_bus_type), WRAP_USB_BUS);
@@ -431,33 +416,22 @@ err_enable:
 NTSTATUS pnp_set_power_state(struct wrap_device *wd,
 			     enum device_power_state state)
 {
-	struct device_object *pdo, *fdo;
-	struct irp *irp;
-	struct io_stack_location *irp_sl;
 	NTSTATUS status;
+	struct device_object *pdo;
+	struct io_stack_location irp_sl;
 
 	pdo = wd->pdo;
-	fdo = IoGetAttachedDevice(pdo);
+	IOTRACE("%p, %p", pdo, IoGetAttachedDevice(pdo));
+	memset(&irp_sl, 0, sizeof(irp_sl));
+	irp_sl.params.power.state.device_state = state;
 	if (state > PowerDeviceD0) {
-		irp = IoAllocateIrp(fdo->stack_count, FALSE);
-		irp_sl = IoGetNextIrpStackLocation(irp);
-		DBGTRACE2("irp = %p, stack = %p", irp, irp_sl);
-		irp_sl->major_fn = IRP_MJ_POWER;
-		irp_sl->minor_fn = IRP_MN_QUERY_POWER;
-		irp_sl->params.power.state.device_state = state;
-		irp->io_status.status = STATUS_NOT_SUPPORTED;
-		status = IoCallDriver(fdo, irp);
+		irp_sl.params.power.state.device_state = state;
+		status = IoSendIrpTopDev(pdo, IRP_MJ_POWER, IRP_MN_QUERY_POWER,
+					 &irp_sl);
 		if (status != STATUS_SUCCESS)
 			WARNING("query power returns %08X", status);
 	}
-	irp = IoAllocateIrp(fdo->stack_count, FALSE);
-	irp_sl = IoGetNextIrpStackLocation(irp);
-	DBGTRACE2("irp = %p, stack = %p", irp, irp_sl);
-	irp_sl->major_fn = IRP_MJ_POWER;
-	irp_sl->minor_fn = IRP_MN_SET_POWER;
-	irp_sl->params.power.state.device_state = state;
-	irp->io_status.status = STATUS_NOT_SUPPORTED;
-	status = IoCallDriver(fdo, irp);
+	status = IoSendIrpTopDev(pdo, IRP_MJ_POWER, IRP_MN_SET_POWER, &irp_sl);
 	if (status != STATUS_SUCCESS)
 		WARNING("set power returns %08X", status);
 	TRACEEXIT1(return status);
@@ -467,26 +441,19 @@ NTSTATUS pnp_start_device(struct wrap_device *wd)
 {
 	struct device_object *fdo;
 	struct device_object *pdo;
-	struct irp *irp;
-	struct io_stack_location *irp_sl;
+	struct io_stack_location irp_sl;
 	NTSTATUS status;
 
 	pdo = wd->pdo;
-	fdo = IoGetAttachedDevice(pdo);
-	DBGTRACE1("fdo: %p, irql: %d", fdo, current_irql());
-	irp = IoAllocateIrp(fdo->stack_count, FALSE);
-	irp_sl = IoGetNextIrpStackLocation(irp);
-	DBGTRACE1("irp = %p, stack = %p", irp, irp_sl);
 	/* TODO: for now we use same resources for both translated
 	 * resources and raw resources */
-	irp_sl->params.start_device.allocated_resources =
+	memset(&irp_sl, 0, sizeof(irp_sl));
+	irp_sl.params.start_device.allocated_resources =
 		wd->resource_list;
-	irp_sl->params.start_device.allocated_resources_translated =
+	irp_sl.params.start_device.allocated_resources_translated =
 		wd->resource_list;
-	irp_sl->major_fn = IRP_MJ_PNP;
-	irp_sl->minor_fn = IRP_MN_START_DEVICE;
-	irp->io_status.status = STATUS_NOT_SUPPORTED;
-	status = IoCallDriver(fdo, irp);
+	status = IoSendIrpTopDev(pdo, IRP_MJ_PNP, IRP_MN_START_DEVICE, &irp_sl);
+	fdo = IoGetAttachedDevice(pdo);
 	if (status == STATUS_SUCCESS)
 		fdo->drv_obj->drv_ext->count++;
 	else
@@ -497,31 +464,18 @@ NTSTATUS pnp_start_device(struct wrap_device *wd)
 
 NTSTATUS pnp_stop_device(struct wrap_device *wd)
 {
-	struct device_object *pdo, *fdo;
-	struct irp *irp;
-	struct io_stack_location *irp_sl;
+	struct device_object *pdo;
 	NTSTATUS status;
 
 	pdo = wd->pdo;
-	fdo = IoGetAttachedDevice(pdo);
-	DBGTRACE1("fdo: %p", fdo);
-	irp = IoAllocateIrp(fdo->stack_count, FALSE);
-	irp_sl = IoGetNextIrpStackLocation(irp);
-	DBGTRACE1("irp = %p, stack = %p", irp, irp_sl);
-	irp_sl->major_fn = IRP_MJ_PNP;
-	irp_sl->minor_fn = IRP_MN_QUERY_STOP_DEVICE;
-	irp->io_status.status = STATUS_NOT_SUPPORTED;
-	status = IoCallDriver(fdo, irp);
+	status = IoSendIrpTopDev(pdo, IRP_MJ_PNP, IRP_MN_QUERY_STOP_DEVICE,
+				 NULL);
 	if (status != STATUS_SUCCESS)
 		WARNING("status: %08X", status);
 	/* for now we ignore query status */
-	irp = IoAllocateIrp(fdo->stack_count, FALSE);
-	irp_sl = IoGetNextIrpStackLocation(irp);
-	DBGTRACE1("irp = %p, stack = %p", irp, irp_sl);
-	irp_sl->major_fn = IRP_MJ_PNP;
-	irp_sl->minor_fn = IRP_MN_STOP_DEVICE;
-	irp->io_status.status = STATUS_NOT_SUPPORTED;
-	status = IoCallDriver(fdo, irp);
+	status = IoSendIrpTopDev(pdo, IRP_MJ_PNP, IRP_MN_STOP_DEVICE, NULL);
+	if (status != STATUS_SUCCESS)
+		WARNING("status: %08X", status);
 	if (status != STATUS_SUCCESS)
 		WARNING("status: %08X", status);
 	TRACEEXIT2(return status);
@@ -531,50 +485,36 @@ NTSTATUS pnp_remove_device(struct wrap_device *wd)
 {
 	struct device_object *pdo, *fdo;
 	struct driver_object *fdo_drv_obj;
-	struct irp *irp;
-	struct io_stack_location *irp_sl;
 	NTSTATUS status;
 
 	pdo = wd->pdo;
 	fdo = IoGetAttachedDevice(pdo);
 	fdo_drv_obj = fdo->drv_obj;
-	DBGTRACE1("fdo: %p", fdo);
-	irp = IoAllocateIrp(fdo->stack_count, FALSE);
-	irp_sl = IoGetNextIrpStackLocation(irp);
-	DBGTRACE1("irp = %p, stack = %p", irp, irp_sl);
-	irp_sl->major_fn = IRP_MJ_PNP;
-	irp_sl->minor_fn = IRP_MN_QUERY_REMOVE_DEVICE;
-	irp->io_status.status = STATUS_NOT_SUPPORTED;
-	status = IoCallDriver(fdo, irp);
-	if (status != STATUS_SUCCESS)
-		WARNING("status: %08X", status);
-	/* for now we ignore query status */
-	irp = IoAllocateIrp(fdo->stack_count, FALSE);
-	irp_sl = IoGetNextIrpStackLocation(irp);
-	DBGTRACE1("irp = %p, stack = %p", irp, irp_sl);
-	irp_sl->major_fn = IRP_MJ_PNP;
-	irp_sl->minor_fn = IRP_MN_REMOVE_DEVICE;
-	irp->io_status.status = STATUS_NOT_SUPPORTED;
-	status = IoCallDriver(fdo, irp);
+	DBGTRACE2("%p, %p, %p", pdo, fdo, fdo_drv_obj);
+	status = IoSendIrpTopDev(pdo, IRP_MJ_PNP, IRP_MN_QUERY_REMOVE_DEVICE,
+				 NULL);
 	if (status != STATUS_SUCCESS)
 		WARNING("status: %08X", status);
 
+	status = IoSendIrpTopDev(pdo, IRP_MJ_PNP, IRP_MN_REMOVE_DEVICE, NULL);
+	if (status != STATUS_SUCCESS)
+		WARNING("status: %08X", status);
 	/* TODO: should we use count in drv_ext or driver's Object
 	 * header reference count to keep count of devices associated
 	 * with a driver? */
 	if (status == STATUS_SUCCESS)
 		fdo_drv_obj->drv_ext->count--;
 	DBGTRACE1("count: %d", fdo_drv_obj->drv_ext->count);
-	if (fdo->drv_obj->drv_ext->count < 0)
+	if (fdo_drv_obj->drv_ext->count < 0)
 		WARNING("wrong count: %d", fdo_drv_obj->drv_ext->count);
 	if (fdo_drv_obj->drv_ext->count == 0) {
 		struct wrap_driver *wrap_driver;
 		DBGTRACE1("unloading driver: %p", fdo_drv_obj);
-		if (fdo_drv_obj->unload)
-			LIN2WIN1(fdo_drv_obj->unload, fdo_drv_obj);
 		wrap_driver =
 			IoGetDriverObjectExtension(fdo_drv_obj,
 					   (void *)CE_WRAP_DRIVER_CLIENT_ID);
+		if (fdo_drv_obj->unload)
+			LIN2WIN1(fdo_drv_obj->unload, fdo_drv_obj);
 		if (wrap_driver) {
 			nt_spin_lock(&loader_lock);
 			unload_wrap_driver(wrap_driver);
@@ -583,6 +523,7 @@ NTSTATUS pnp_remove_device(struct wrap_device *wd)
 			ERROR("couldn't get wrap_driver");
 		ObDereferenceObject(fdo_drv_obj);
 	}
+	IoDeleteDevice(pdo);
 	TRACEEXIT1(return status);
 }
 
@@ -644,8 +585,12 @@ static int wrap_pnp_start_device(struct wrap_device *wd)
 		IoDeleteDevice(pdo);
 		return -EINVAL;
 	}
-	if (pnp_start_device(wd) != STATUS_SUCCESS)
+	if (pnp_start_device(wd) != STATUS_SUCCESS) {
+		/* TODO: we need proper cleanup, to deallocate memory,
+		 * for example */
+//		pnp_remove_device(wd);
 		return -EINVAL;
+	}
 	return 0;
 }
 
@@ -670,9 +615,8 @@ void __devexit wrap_pnp_remove_pci_device(struct pci_dev *pdev)
 {
 	struct wrap_device *wd;
 
-	TRACEENTER1("%p", pdev);
 	wd = (struct wrap_device *)pci_get_drvdata(pdev);
-	TRACEENTER1("%p", wd);
+	TRACEENTER1("%p, %p", pdev, wd);
 	if (!wd)
 		TRACEEXIT1(return);
 	pnp_remove_device(wd);
@@ -707,13 +651,12 @@ void *wrap_pnp_start_usb_device(struct usb_device *udev,
 	struct wrap_device *wd;
 	int ret;
 	wd = &wrap_devices[usb_id->driver_info];
-	DBGTRACE1("%04x, %04x, %x, %x, %x, %x, %x, %x, %x, %p",
-		  usb_id->idVendor, usb_id->idProduct,
-		  usb_id->bDeviceClass, usb_id->bDeviceSubClass,
-		  usb_id->bDeviceProtocol, usb_id->bInterfaceClass,
-		  usb_id->bDeviceProtocol, usb_id->bInterfaceClass,
-		  usb_id->bInterfaceSubClass, wd->usb.intf);
-
+	TRACEENTER1("%04x, %04x, %x, %x, %x, %x, %x, %x, %x, %p",
+		    usb_id->idVendor, usb_id->idProduct,
+		    usb_id->bDeviceClass, usb_id->bDeviceSubClass,
+		    usb_id->bDeviceProtocol, usb_id->bInterfaceClass,
+		    usb_id->bDeviceProtocol, usb_id->bInterfaceClass,
+		    usb_id->bInterfaceSubClass, wd->usb.intf);
 	/* TI 4150 needs a full reset */
 	if (usb_id->idVendor == 0x057c && usb_id->idProduct == 0x5601 &&
 	    (xchg(&wd->usb.reset, 1) == 0)) {
@@ -728,11 +671,13 @@ void *wrap_pnp_start_usb_device(struct usb_device *udev,
 			goto out;
 		}
 	}
-	/* USB device may have multiple interfaces; initialize a
-	  device only once */
+	/* USB device (e.g., RNDIS) may have multiple interfaces;
+	  initialize one interface only (is there a way to know which
+	  of these interfaces is for network?) */
 	if (wd->usb.intf) {
 		DBGTRACE1("device already initialized: %p", wd->usb.intf);
-		ret = 0;
+		usb_set_intfdata(intf, NULL);
+		TRACEEXIT1(return -EINVAL);
 	} else {
 #if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
 		wd->usb.udev = interface_to_usbdev(intf);
@@ -748,9 +693,10 @@ void *wrap_pnp_start_usb_device(struct usb_device *udev,
 out:
 	DBGTRACE2("ret: %d", ret);
 #if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
-	if (ret)
-		return -EINVAL;
-	else
+	if (ret) {
+		wd->usb.intf = NULL;
+		TRACEEXIT1(return -EINVAL);
+	} else
 		return 0;
 #else
 	if (ret)
@@ -765,8 +711,8 @@ void wrap_pnp_remove_usb_device(struct usb_interface *intf)
 {
 	struct wrap_device *wd;
 
-	TRACEENTER1("%p", intf);
 	wd = (struct wrap_device *)usb_get_intfdata(intf);
+	TRACEENTER1("%p, %p", intf, wd);
 	if (wd == NULL)
 		TRACEEXIT1(return);
 	usb_set_intfdata(intf, NULL);
@@ -782,8 +728,7 @@ void wrap_pnp_remove_usb_device(struct usb_device *udev, void *ptr)
 	struct wrap_device *wd = ptr;
 	struct usb_interface *intf;
 
-	TRACEENTER1("%p", wd);
-
+	TRACEENTER1("%p, %p", intf, wd);
 	if (wd == NULL)
 		TRACEEXIT1(return);
 	intf = wd->usb.intf;
@@ -800,29 +745,25 @@ int wrap_pnp_suspend_usb_device(struct usb_interface *intf, pm_message_t state)
 	struct device_object *pdo;
 
 	wd = usb_get_intfdata(intf);
+	TRACEENTER1("%p, %p", intf, wd);
 	if (!wd)
-		TRACEEXIT1(return -1);
+		TRACEEXIT1(return 0);
 	pdo = wd->pdo;
 	if (pnp_set_power_state(wd, PowerDeviceD3))
 		return -1;
-	else {
-//		intf->dev.power.power_state.event = state.event;
-		return 0;
-	}
+	return 0;
 }
 
 int wrap_pnp_resume_usb_device(struct usb_interface *intf)
 {
 	struct wrap_device *wd;
 	wd = usb_get_intfdata(intf);
+	TRACEENTER1("%p, %p", intf, wd);
 	if (!wd)
-		TRACEEXIT1(return -1);
+		TRACEEXIT1(return 0);
 	if (pnp_set_power_state(wd, PowerDeviceD0))
 		return -1;
-	else {
-//		intf->dev.power.power_state.event = PM_EVENT_ON;
-		return 0;
-	}
+	return 0;
 }
 #endif
 
