@@ -103,30 +103,25 @@ static int procfs_read_ndis_encr(char *page, char **start, off_t off,
 		essid.essid[essid.length] = '\0';
 		p += sprintf(p, "essid=%s\n", essid.essid);
 	}
-
 	res = miniport_query_int(wnd, OID_802_11_ENCRYPTION_STATUS,
 				 &encr_status);
-	res |= miniport_query_int(wnd, OID_802_11_AUTHENTICATION_MODE,
-				  &auth_mode);
-
 	if (!res) {
-		int t = wnd->encr_info.tx_key_index;
+		typeof(&wnd->encr_info.keys[0]) tx_key;
 		p += sprintf(p, "tx_key=%u\n", wnd->encr_info.tx_key_index);
 		p += sprintf(p, "key=");
-		if (wnd->encr_info.keys[t].length > 0)
-			for (i = 0; i < NDIS_ENCODING_TOKEN_MAX &&
-				     i < wnd->encr_info.keys[t].length;
-			     i++)
-				p += sprintf(p, "%2.2X",
-					     wnd->encr_info.keys[t].key[i]);
+		tx_key = &wnd->encr_info.keys[wnd->encr_info.tx_key_index];
+		if (tx_key->length > 0)
+			for (i = 0; i < tx_key->length; i++)
+				p += sprintf(p, "%2.2X", tx_key->key[i]);
 		else
 			p += sprintf(p, "off");
 		p += sprintf(p, "\n");
-
 		p += sprintf(p, "encr_mode=%d\n", encr_status);
-		p += sprintf(p, "auth_mode=%d\n", auth_mode);
 	}
-
+	res = miniport_query_int(wnd, OID_802_11_AUTHENTICATION_MODE,
+				  &auth_mode);
+	if (!res)
+		p += sprintf(p, "auth_mode=%d\n", auth_mode);
 	res = miniport_query_int(wnd, OID_802_11_INFRASTRUCTURE_MODE,
 				 &infra_mode);
 	p += sprintf(p, "mode=%s\n", (infra_mode == Ndis802_11IBSS) ?
@@ -154,6 +149,7 @@ static int procfs_read_ndis_hw(char *page, char **start, off_t off,
 	ndis_rts_threshold rts_threshold;
 	ndis_fragmentation_threshold frag_threshold;
 	ndis_antenna antenna;
+	ULONG packet_filter;
 
 	if (off != 0) {
 		*eof = 1;
@@ -198,8 +194,7 @@ static int procfs_read_ndis_hw(char *page, char **start, off_t off,
 	res = miniport_query_int(wnd, OID_802_11_POWER_MODE, &power_mode);
 	if (!res)
 		p += sprintf(p, "power_mode=%s\n",
-			     (power_mode == NDIS_POWER_OFF) ?
-			     "always_on" :
+			     (power_mode == NDIS_POWER_OFF) ? "always_on" :
 			     (power_mode == NDIS_POWER_MAX) ?
 			     "max_savings" : "min_savings");
 
@@ -236,6 +231,14 @@ static int procfs_read_ndis_hw(char *page, char **start, off_t off,
 		     test_bit(Ndis802_11AuthModeWPA2PSK, &wnd->capa.auth) ?
 		     ", WPA2PSK" : "");
 
+	res = miniport_query_int(wnd, OID_GEN_CURRENT_PACKET_FILTER,
+				 &packet_filter);
+	if (!res) {
+		if (packet_filter != wnd->packet_filter)
+			WARNING("wrong packet_filter? 0x%08x, 0x%08x\n",
+				packet_filter, wnd->packet_filter);
+		p += sprintf(p, "packet_filter: 0x%08x\n", packet_filter);
+	}
 	if (p - page > count) {
 		WARNING("wrote %lu bytes (limit is %u)",
 			(unsigned long)(p - page), count);
@@ -273,6 +276,8 @@ static int procfs_write_ndis_settings(struct file *file, const char *buf,
 {
 	struct wrap_ndis_device *wnd = (struct wrap_ndis_device *)data;
 	char setting[MAX_PROC_STR_LEN], *p;
+	unsigned int i;
+	NTSTATUS res;
 
 	if (count > MAX_PROC_STR_LEN)
 		return -EINVAL;
@@ -288,8 +293,6 @@ static int procfs_write_ndis_settings(struct file *file, const char *buf,
 		*p = 0;
 
 	if (!strcmp(setting, "hangcheck_interval")) {
-		int i;
-
 		if (!p)
 			return -EINVAL;
 		p++;
@@ -297,33 +300,7 @@ static int procfs_write_ndis_settings(struct file *file, const char *buf,
 		hangcheck_interval = i;
 		hangcheck_del(wnd);
 		hangcheck_add(wnd);
-	} else if (!strcmp(setting, "check_capa")) {
-		check_capa(wnd);
-		printk(KERN_INFO
-		       "%s: encryption modes supported: %s%s%s%s%s%s%s\n",
-		       wnd->net_dev->name,
-		       test_bit(Ndis802_11Encryption1Enabled,
-				&wnd->capa.encr) ?
-		       "WEP" : "none",
-
-		       test_bit(Ndis802_11Encryption2Enabled,
-				&wnd->capa.encr) ?
-		       "; TKIP with WPA" : "",
-		       test_bit(Ndis802_11AuthModeWPA2, &wnd->capa.auth) ?
-		       ", WPA2" : "",
-		       test_bit(Ndis802_11AuthModeWPA2PSK, &wnd->capa.auth) ?
-		       ", WPA2PSK" : "",
-
-		       test_bit(Ndis802_11Encryption3Enabled,
-				&wnd->capa.encr) ?
-		       "; AES/CCMP with WPA" : "",
-		       test_bit(Ndis802_11AuthModeWPA2, &wnd->capa.auth) ?
-		       ", WPA2" : "",
-		       test_bit(Ndis802_11AuthModeWPA2PSK, &wnd->capa.auth) ?
-		       ", WPA2PSK" : "");
 	} else if (!strcmp(setting, "suspend")) {
-		int i;
-
 		if (!p)
 			return -EINVAL;
 		p++;
@@ -343,7 +320,6 @@ static int procfs_write_ndis_settings(struct file *file, const char *buf,
 		if (i)
 			return -EINVAL;
 	} else if (!strcmp(setting, "resume")) {
-		int i;
 		if (wrap_is_pci_bus(wnd->wd->dev_bus_type))
 			i = wrap_pnp_resume_pci_device(wnd->wd->pci.pdev);
 		else
@@ -354,87 +330,7 @@ static int procfs_write_ndis_settings(struct file *file, const char *buf,
 #endif
 		if (i)
 			return -EINVAL;
-	} else if (!strcmp(setting, "reinit")) {
-		if (ndis_reinit(wnd))
-			return -EINVAL;
-#ifdef USB_DEBUG
-	} else if (!strcmp(setting, "irp")) {
-		struct irp *irp;
-		struct io_stack_location *irp_sl;
-		NTSTATUS status;
-		int major_fn, minor_fn, n;
-		struct device_object *dev;
-		if (!p)
-			return -EINVAL;
-		p++;
-		n = sscanf(p, "%d,%d,%x", &major_fn, &minor_fn, (int *)&dev);
-		DBGTRACE1("n = %d, mj = %d, mn = %d, dev = %p", n, major_fn,
-			  minor_fn, dev);
-		if (n != 3)
-			return -EINVAL;
-		irp = IoAllocateIrp(dev->stack_count, FALSE);
-		DBGTRACE1("count: %d, irp = %p", dev->stack_count, irp);
-		DBGTRACE1("drv_obj: %p", dev->drv_obj);
-		irp_sl = IoGetNextIrpStackLocation(irp);
-		irp_sl->major_fn = major_fn;
-		irp_sl->minor_fn = minor_fn;
-		irp->io_status.status = STATUS_NOT_SUPPORTED;
-		status = 0;
-		status = IoCallDriver(dev, irp);
-		DBGTRACE1("status = %d", status);
-#endif
-	} else if (!strcmp(setting, "power_profile")) {
-		int i;
-		struct miniport_char *miniport;
-		ULONG profile_inf;
-
-		if (!p)
-			return -EINVAL;
-		p++;
-		i = simple_strtol(p, NULL, 10);
-		if (i < 0 || i > 1)
-			return -EINVAL;
-
-		miniport = &wnd->wd->driver->ndis_driver->miniport;
-		if (!miniport->pnp_event_notify)
-			return -EFAULT;
-
-		/* 1 for AC and 0 for Battery */
-		if (i)
-			profile_inf = NdisPowerProfileAcOnLine;
-		else
-			profile_inf = NdisPowerProfileBattery;
-
-		miniport->pnp_event_notify(wnd->nmb->adapter_ctx,
-					   NdisDevicePnPEventPowerProfileChanged,
-					   &profile_inf, sizeof(profile_inf));
-	} else if (!strcmp(setting, "auth_mode")) {
-		int i;
-
-		if (!p)
-			return -EINVAL;
-		p++;
-		i = simple_strtol(p, NULL, 10);
-		if (i <= 0 || i > 5)
-			return -EINVAL;
-
-		if (set_auth_mode(wnd, i))
-			return -EINVAL;
-	} else if (!strcmp(setting, "encr_mode")) {
-		int i;
-
-		if (!p)
-			return -EINVAL;
-		p++;
-		i = simple_strtol(p, NULL, 10);
-		if (i <= 0 || i > 7)
-			return -EINVAL;
-
-		if (set_encr_mode(wnd, i))
-			return -EINVAL;
 	} else if (!strcmp(setting, "stats_enabled")) {
-		int i;
-
 		if (!p)
 			return -EINVAL;
 		p++;
@@ -443,44 +339,14 @@ static int procfs_write_ndis_settings(struct file *file, const char *buf,
 			wnd->stats_enabled = TRUE;
 		else
 			wnd->stats_enabled = FALSE;
-	} else if (!strcmp(setting, "tx_antenna")) {
-		ndis_antenna antenna;
-		unsigned int i;
-		NTSTATUS res;
-
+	} else if (!strcmp(setting, "packet_filter")) {
 		if (!p)
 			return -EINVAL;
 		p++;
-		i = simple_strtol(p, NULL, 16);
-		res = miniport_query_info(wnd, OID_802_11_NUMBER_OF_ANTENNAS,
-					  &antenna, sizeof(antenna));
+		i = simple_strtol(p, NULL, 10);
+		res = miniport_set_int(wnd, OID_GEN_CURRENT_PACKET_FILTER, i);
 		if (res)
-			return -EINVAL;
-		if (i >= -1 && i < antenna)
-			antenna = i;
-		res = miniport_set_info(wnd, OID_802_11_TX_ANTENNA_SELECTED,
-				  &antenna, sizeof(antenna));
-		if (res)
-			return -EINVAL;
-	} else if (!strcmp(setting, "rx_antenna")) {
-		ndis_antenna antenna;
-		unsigned int i;
-		NTSTATUS res;
-
-		if (!p)
-			return -EINVAL;
-		p++;
-		i = simple_strtol(p, NULL, 16);
-		res = miniport_query_info(wnd, OID_802_11_NUMBER_OF_ANTENNAS,
-					  &antenna, sizeof(antenna));
-		if (res)
-			return -EINVAL;
-		if (i >= -1 && i < antenna)
-			antenna = i;
-		res = miniport_set_info(wnd, OID_802_11_RX_ANTENNA_SELECTED,
-				  &antenna, sizeof(antenna));
-		if (res)
-			return -EINVAL;
+			WARNING("setting packet_filter failed: %08X", res);
 	}
 	return count;
 }
@@ -577,12 +443,19 @@ static int procfs_read_debug(char *page, char **start, off_t off,
 			     int count, int *eof, void *data)
 {
 	char *p = page;
+	enum alloc_type type;
 
 	if (off != 0) {
 		*eof = 1;
 		return 0;
 	}
 	p += sprintf(p, "%d\n", debug);
+	type = 0;
+#ifdef ALLOC_INFO
+	for (type = 0; type < ALLOC_TYPE_MAX; type++)
+		p += sprintf(p, "total size of allocations in %d: %d\n",
+			     type, alloc_size(type));
+#endif
 	return (p - page);
 }
 
