@@ -21,12 +21,12 @@
 static unsigned int urb_id = 0;
 
 #define DUMP_WRAP_URB(wrap_urb, dir)					\
-	USBTRACE("urb %p (%d) %s: buf: %p, len: %d, pipe: 0x%x",	\
+	USBTRACE("urb %p (%d) %s: buf: %p, len: %d, pipe: 0x%x, %d",	\
 		 (wrap_urb)->urb, (wrap_urb)->id,			\
 		 (dir == USB_DIR_OUT) ? "going down" : "coming back",	\
 		 (wrap_urb)->urb->transfer_buffer,			\
 		 (wrap_urb)->urb->transfer_buffer_length,		\
-		 (wrap_urb)->urb->pipe)
+		 (wrap_urb)->urb->pipe, (wrap_urb)->urb->status)
 
 #define DUMP_URB_BUFFER(urb, dir)					\
 	while (debug >= 2) {						\
@@ -116,7 +116,7 @@ static STDCALL void wrap_cancel_irp(struct device_object *dev_obj,
 				    struct irp *irp);
 
 /* use tasklet instead worker to process completed urbs */
-//#define USB_TASKLET 1
+#define USB_TASKLET 1
 
 #ifdef USB_TASKLET
 static struct tasklet_struct wrap_urb_complete_work;
@@ -126,6 +126,8 @@ static struct work_struct wrap_urb_complete_work;
 static void wrap_urb_complete_worker(void *dummy);
 #endif
 
+static void wrap_free_urb(struct urb *urb);
+
 int usb_init(void)
 {
 	InitializeListHead(&wrap_urb_complete_list);
@@ -167,19 +169,22 @@ void usb_exit_device(struct wrap_device *wd)
 	KIRQL irql;
 
 	USBENTER("%d", wd->usb.num_alloc_urbs);
-	IoAcquireCancelSpinLock(&irql);
-	while ((ent = RemoveHeadList(&wd->usb.wrap_urb_list))) {
+	while (1) {
+		IoAcquireCancelSpinLock(&irql);
+		ent = RemoveHeadList(&wd->usb.wrap_urb_list);
+		IoReleaseCancelSpinLock(irql);
+		if (!ent)
+			break;
 		wrap_urb = container_of(ent, struct wrap_urb, list);
 		if (wrap_urb->state == URB_SUBMITTED) {
 			WARNING("Windows driver %s didn't free urb: %p",
 				wd->driver->name, wrap_urb->urb);
-			wrap_cancel_urb(wrap_urb);
-		} else {
-			usb_free_urb(wrap_urb->urb);
-			kfree(wrap_urb);
+			usb_kill_urb(wrap_urb->urb);
 		}
+		usb_free_urb(wrap_urb->urb);
+		kfree(wrap_urb);
 	}
-	IoReleaseCancelSpinLock(irql);
+	wd->usb.num_alloc_urbs = 0;
 	USBEXIT(return);
 }
 
@@ -272,13 +277,13 @@ static void wrap_free_urb(struct urb *urb)
 void wrap_suspend_urbs(struct wrap_device *wd)
 {
 	/* TODO: do we need to cancel urbs? */
-	USBTRACE("%p", wd);
+	USBTRACE("%p, %d", wd, wd->usb.num_alloc_urbs);
 }
 
 void wrap_resume_urbs(struct wrap_device *wd)
 {
 	/* TODO: do we need to resubmit urbs? */
-	USBTRACE("%p", wd);
+	USBTRACE("%p, %d", wd, wd->usb.num_alloc_urbs);
 }
 
 static struct urb *wrap_alloc_urb(struct irp *irp, unsigned int pipe,
@@ -402,6 +407,7 @@ static USBD_STATUS wrap_submit_urb(struct irp *irp)
 	NT_URB_STATUS(nt_urb) = USBD_STATUS_PENDING;
 	IoMarkIrpPending(irp);
 	DUMP_URB_BUFFER(urb, USB_DIR_OUT);
+	USBTRACE("%p", urb);
 #if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
 	ret = usb_submit_urb(urb, alloc_flags);
 #else
@@ -412,7 +418,7 @@ static USBD_STATUS wrap_submit_urb(struct irp *irp)
 		wrap_free_urb(urb);
 		/* we assume that IRP was not in pending state before */
 		IoUnmarkIrpPending(irp);
-		NT_URB_STATUS(nt_urb) = USBD_STATUS_REQUEST_FAILED;
+		NT_URB_STATUS(nt_urb) = wrap_urb_status(ret);
 		USBEXIT(return NT_URB_STATUS(nt_urb));
 	} else
 		USBEXIT(return USBD_STATUS_PENDING);
@@ -1148,6 +1154,8 @@ NTSTATUS wrap_submit_irp(struct device_object *pdo, struct irp *irp)
 
 	USBTRACE("%p, %p", pdo, irp);
 	wd = pdo->reserved;
+	if (wd->usb.intf == NULL)
+		USBEXIT(return STATUS_DEVICE_REMOVED);
 	irp->wd = wd;
 	irp_sl = IoGetCurrentIrpStackLocation(irp);
 	switch (irp_sl->params.dev_ioctl.code) {
@@ -1265,9 +1273,9 @@ WRAP_EXPORT(USBD_ParseConfigurationDescriptorEx)
 	struct usb_interface_descriptor *intf;
 
 	USBENTER("config = %p, start = %p, ifnum = %d, alt_setting = %d,"
-		      " class = %d, subclass = %d, proto = %d", config, start,
-		      bInterfaceNumber, bAlternateSetting,
-		      bInterfaceClass, bInterfaceSubClass, bInterfaceProtocol);
+		 " class = %d, subclass = %d, proto = %d", config, start,
+		 bInterfaceNumber, bAlternateSetting, bInterfaceClass,
+		 bInterfaceSubClass, bInterfaceProtocol);
 
 	for (pos = start; pos < ((void *)config + config->wTotalLength);
 	     pos += intf->bLength) {
