@@ -138,6 +138,7 @@
 
 #define NOREGPARM __attribute__((regparm(0)))
 #define packed __attribute__((packed))
+#define __unused __attribute((unused))
 
 typedef u8	BOOLEAN;
 typedef u8	BYTE;
@@ -230,8 +231,10 @@ typedef STDCALL void (*DPC)(struct kdpc *kdpc, void *ctx, void *arg1,
 
 struct kdpc {
 	SHORT type;
-	/* number is used to represent if the dpc is queued or not */
-	UCHAR number;
+	/* we don't implement kdpc executing on a specific processor;
+	 * until then number is used to represent if the dpc is queued
+	 * or not */
+	UCHAR nr_cpu;
 	UCHAR importance;
 	struct nt_list list;
 
@@ -270,9 +273,7 @@ enum lock_operation {
 };
 
 enum mode {
-	KernelMode,
-	UserMode,
-	MaximumMode
+	KernelMode, UserMode, MaximumMode
 };
 
 struct mdl {
@@ -302,8 +303,11 @@ struct mdl {
 #define MDL_ALLOCATED_MUST_SUCCEED	0x4000
 #define MDL_CACHE_ALLOCATED		0x8000
 
+#define page_start(ptr) ((void *)((ULONG_PTR)(ptr) & ~(PAGE_SIZE - 1)))
+#define byte_offset(ptr) ((ULONG)((ULONG_PTR)(ptr) & (PAGE_SIZE - 1)))
+
 #define MmGetMdlByteCount(mdl) ((mdl)->bytecount)
-#define MmGetMdlVirtualAddress(mdl) ((mdl)->startva)
+#define MmGetMdlVirtualAddress(mdl) ((mdl)->startva + (mdl)->byteoffset)
 #define MmGetMdlByteOffset(mdl) ((mdl)->byteoffset)
 #define MmGetSystemAddressForMdl(mdl) ((mdl)->mappedsystemva)
 #define MmGetSystemAddressForMdlSafe(mdl, priority) ((mdl)->mappedsystemva)
@@ -313,9 +317,10 @@ do {									\
 	(mdl)->next = NULL;						\
 	(mdl)->size = MmSizeOfMdl(baseva, length);			\
 	(mdl)->flags = 0;						\
-	(mdl)->startva = baseva;					\
-	(mdl)->byteoffset = (ULONG)offset_in_page(baseva);		\
+	(mdl)->startva = page_start(baseva);				\
+	(mdl)->byteoffset = byte_offset(baseva);			\
 	(mdl)->bytecount = length;					\
+	(mdl)->mappedsystemva = baseva;					\
 	DBGTRACE4("%p %p %p %d %d", (mdl), baseva, (mdl)->startva,	\
 		  (mdl)->byteoffset, length);				\
 } while (0)
@@ -348,7 +353,7 @@ struct wait_block {
 	struct nt_list list;
 	struct task_struct *thread;
 	void *object;
-	void *event_wait;
+	void *thread_wait;
 	USHORT wait_key;
 	USHORT wait_type;
 };
@@ -490,6 +495,7 @@ struct dev_obj_ext {
 	CSHORT type;
 	CSHORT size;
 	struct device_object *dev_obj;
+	struct device_object *attached_to;
 };
 
 struct io_status_block {
